
import numpy as np
import sympy as sym

def gfpizv(grid,data,name,bus_name):

    # inputs
    xi_freq = sym.Symbol('xi_freq',real=True)
    omega_coi = sym.Symbol('omega_coi',real=True)

    # parameters
    S_n,U_n = sym.symbols(f'S_n_{name},U_n_{name}', real=True)
    T_e = sym.Symbol(f'H_{name}', real=True)
    K_f = sym.Symbol(f'K_f_{name}', real=True)
    T_f = sym.Symbol(f'T_f_{name}', real=True)
    K_agc = sym.Symbol(f'K_agc_{name}', real=True)
    T_e = sym.Symbol(f'T_e_{name}', real=True)
    K_p = sym.Symbol(f"K_p_{name}", real=True)    
    T_p = sym.Symbol(f"T_p_{name}", real=True)   
    T_c = sym.Symbol(f"T_c_{name}", real=True)   
    T_v = sym.Symbol(f"T_v_{name}", real=True)   
    K_qp = sym.Symbol(f"K_qp_{name}", real=True)   
    K_qi = sym.Symbol(f"K_qi_{name}", real=True) 
    Droop = sym.Symbol(f"Droop_{name}", real=True)   
    R_v,X_v = sym.symbols(f'R_v_{name},X_v_{name}', real=True)
    K_delta = sym.Symbol(f"K_delta_{name}", real=True) 
    K_soc = sym.Symbol(f"K_soc_{name}", real=True) 

    ## inputs
    omega_ref = sym.Symbol(f'omega_{name}_ref', real=True)
    p_c = sym.Symbol(f'p_c_{name}', real=True)  
    q_ref = sym.Symbol(f'q_ref_{name}', real=True)    
    e_ao_m,e_bo_m,e_co_m,e_no_m = sym.symbols(f'e_ao_m_{name},e_bo_m_{name},e_co_m_{name},e_no_m_{name}', real=True)
    phi_a = sym.Symbol(f'phi_a_{name}', real=True)
    phi_b = sym.Symbol(f'phi_b_{name}', real=True)
    phi_c = sym.Symbol(f'phi_c_{name}', real=True)
    phi_n = sym.Symbol(f'phi_n_{name}', real=True)
    v_ra = sym.Symbol(f'v_ra_{name}', real=True)
    v_rb = sym.Symbol(f'v_rb_{name}', real=True)
    v_rc = sym.Symbol(f'v_rc_{name}', real=True)
    v_rn = sym.Symbol(f'v_rn_{name}', real=True)    
    v_sa_r,v_sb_r,v_sc_r,v_sn_r,v_og_r = sym.symbols(f'V_{bus_name}_0_r,V_{bus_name}_1_r,V_{bus_name}_2_r,V_{bus_name}_3_r,v_{bus_name}_o_r', real=True)
    v_sa_i,v_sb_i,v_sc_i,v_sn_i,v_og_i = sym.symbols(f'V_{bus_name}_0_i,V_{bus_name}_1_i,V_{bus_name}_2_i,V_{bus_name}_3_i,v_{bus_name}_o_i', real=True)
    i_sa_r,i_sb_r,i_sc_r,i_sn_r,i_ng_r = sym.symbols(f'i_vsc_{name}_a_r,i_vsc_{name}_b_r,i_vsc_{name}_c_r,i_vsc_{name}_n_r,i_vsc_{name}_ng_r', real=True)
    i_sa_i,i_sb_i,i_sc_i,i_sn_i,i_ng_i = sym.symbols(f'i_vsc_{name}_a_i,i_vsc_{name}_b_i,i_vsc_{name}_c_i,i_vsc_{name}_n_i,i_vsc_{name}_ng_i', real=True)
    v_ta_r,v_tb_r,v_tc_r,v_tn_r = sym.symbols(f'v_ta_r_{name},v_tb_r_{name},v_tc_r_{name},v_tn_r_{name}', real=True)
    v_ta_i,v_tb_i,v_tc_i,v_tn_i = sym.symbols(f'v_ta_i_{name},v_tb_i_{name},v_tc_i_{name},v_tn_i_{name}', real=True)
    p_soc = sym.Symbol(f'p_soc_{name}',real=True)
    v_dc = sym.Symbol(f'v_dc_{name}',real=True)


    # dynamical states
    p_cf= sym.Symbol(f'p_cf_{name}', real=True)   
    phi = sym.Symbol(f'phi_{name}', real=True)
    xi_p = sym.Symbol(f'xi_p_{name}', real=True)
    xi_q = sym.Symbol(f'xi_q_{name}', real=True)
    p_ef = sym.Symbol(f'p_ef_{name}', real=True)
    De_ao_m,De_bo_m,De_co_m,De_no_m  = sym.symbols(f'De_ao_m_{name},De_bo_m_{name},De_co_m_{name},De_no_m_{name}', real=True)
    omega = sym.Symbol(f'omega_{name}', real=True)
    
    ## auxiliar variables
    omega = sym.Symbol(f'omega_{name}', real=True)
    p_m = sym.Symbol(f'p_m_{name}', real=True) 

    
    ## auxiliar equations
    alpha = np.exp(2.0/3*np.pi*1j)
    A_0a =  np.array([[1, 1, 1],
                    [1, alpha**2, alpha],
                    [1, alpha, alpha**2]])

    A_a0 = 1/3* np.array([[1, 1, 1],
                        [1, alpha, alpha**2],
                        [1, alpha**2, alpha]])

    omega_coi_i = 0
    HS_coi = 0

    ## POI currents
    i_sa = i_sa_r + 1j*i_sa_i
    i_sb = i_sb_r + 1j*i_sb_i
    i_sc = i_sc_r + 1j*i_sc_i
    i_sn = i_sn_r + 1j*i_sn_i

    ## POI voltages
    v_sa = v_sa_r + 1j*v_sa_i
    v_sb = v_sb_r + 1j*v_sb_i
    v_sc = v_sc_r + 1j*v_sc_i
    v_sn = v_sn_r + 1j*v_sn_i
    v_og = v_og_r + 1j*v_og_i

    v_sabc = sym.Matrix([[v_sa],[v_sb],[v_sc]])
    i_sabc = sym.Matrix([[i_sa],[i_sb],[i_sc]])

    v_szpn = A_a0*v_sabc
    i_szpn = A_a0*i_sabc
    
    s_pos = 3*v_szpn[1]*sym.conjugate(i_szpn[1])
    s_neg = 3*v_szpn[2]*sym.conjugate(i_szpn[2])
    s_zer = 3*v_szpn[0]*sym.conjugate(i_szpn[0])
    
    p_pos = sym.re(s_pos)
    q_pos = sym.im(s_pos)

    p_r = K_agc*xi_freq
    epsilon_p = p_m - p_ef
    epsilon_q = q_ref - q_pos/S_n
    De_q_nosat = K_qp*epsilon_q + K_qi*xi_q
    De_min = -U_n*0.05
    De_max =  U_n*0.05
    De_q  = sym.Piecewise((De_min,De_q_nosat<De_min),
                          (De_max,De_q_nosat>De_max),
                          (De_q_nosat,True))
    K_qaw = sym.Piecewise((0,De_q_nosat<De_min),
                          (0,De_q_nosat>De_max),
                          (1,True))
    Z_va = R_v + 1j*X_v # virtual impedance phase a
    Z_vb = R_v + 1j*X_v # virtual impedance phase b
    Z_vc = R_v + 1j*X_v # virtual impedance phase c

    e_ao_r = (e_ao_m+De_ao_m)*sym.cos(phi_a + phi) 
    e_ao_i = (e_ao_m+De_ao_m)*sym.sin(phi_a + phi) 
    e_bo_r = (e_bo_m+De_bo_m)*sym.cos(phi_b + phi-2/3*np.pi) 
    e_bo_i = (e_bo_m+De_bo_m)*sym.sin(phi_b + phi-2/3*np.pi) 
    e_co_r = (e_co_m+De_co_m)*sym.cos(phi_c + phi-4/3*np.pi) 
    e_co_i = (e_co_m+De_co_m)*sym.sin(phi_c + phi-4/3*np.pi) 
    e_no_r = (e_no_m+De_no_m)*sym.cos(phi_n + phi) 
    e_no_i = (e_no_m+De_no_m)*sym.sin(phi_n + phi)

    e_ao_cplx = e_ao_r + 1j*e_ao_i
    e_bo_cplx = e_bo_r + 1j*e_bo_i
    e_co_cplx = e_co_r + 1j*e_co_i
    e_no_cplx = e_no_r + 1j*e_no_i

    ## VSC AC side voltages
    v_ta = v_ta_r + 1j*v_ta_i
    v_tb = v_tb_r + 1j*v_tb_i
    v_tc = v_tc_r + 1j*v_tc_i
    v_tn = v_tn_r + 1j*v_tn_i

    ## differential equations
    dphi   = 2*np.pi*50*(omega - omega_coi) - K_delta*phi
    dxi_p = epsilon_p
    dxi_q = K_qaw*epsilon_q - (1-K_qaw)*xi_q  - 1e-6*xi_q
    dp_ef = 1/T_e*(p_pos/S_n - p_ef)
    dp_cf = 1/T_c*(p_c - p_cf)
    dDe_ao_m = 1/T_v*(v_ra + De_q - De_ao_m)
    dDe_bo_m = 1/T_v*(v_rb + De_q - De_bo_m)
    dDe_co_m = 1/T_v*(v_rc + De_q - De_co_m)
    dDe_no_m = 1/T_v*(v_rn - De_no_m)

    grid.dae['f'] += [dphi,dxi_p,dxi_q,dp_ef,dp_cf,dDe_ao_m,dDe_bo_m,dDe_co_m,dDe_no_m] #,dDe_ao_m,dDe_bo_m,dDe_co_m,dDe_no_m]
    grid.dae['x'] += [ phi, xi_p, xi_q, p_ef, p_cf, De_ao_m, De_bo_m, De_co_m, De_no_m] #, De_ao_m, De_bo_m, De_co_m, De_no_m]
    # grid.dae['f'] += [dphi,dxi_p,dp_ef,dp_cf,dDe_ao_m,dDe_bo_m,dDe_co_m,dDe_no_m] #,dDe_ao_m,dDe_bo_m,dDe_co_m,dDe_no_m]
    # grid.dae['x'] += [ phi, xi_p, p_ef, p_cf, De_ao_m, De_bo_m, De_co_m, De_no_m] #, De_ao_m, De_bo_m, De_co_m, De_no_m]
        
    ## algebraic equations   
    g_omega = -omega + K_p*(epsilon_p + xi_p/T_p) + 1
    g_p_m  = -p_m + p_cf + p_r - 1/Droop*(omega - omega_ref)  + K_soc*p_soc
    eq_v_ta_cplx =  e_ao_cplx - i_sa*Z_va - v_ta   # v_sa = v_sag
    eq_v_tb_cplx =  e_bo_cplx - i_sb*Z_vb - v_tb
    eq_v_tc_cplx =  e_co_cplx - i_sc*Z_vc - v_tc

    grid.dae['g'] +=     [g_omega, g_p_m] 
    grid.dae['y_ini'] += [  omega,   p_m] 
    grid.dae['y_run'] += [  omega,   p_m] 
 
    grid.dae['g'] += [sym.re(eq_v_ta_cplx)] 
    grid.dae['g'] += [sym.re(eq_v_tb_cplx)] 
    grid.dae['g'] += [sym.re(eq_v_tc_cplx)] 
    grid.dae['g'] += [sym.im(eq_v_ta_cplx)]
    grid.dae['g'] += [sym.im(eq_v_tb_cplx)]
    grid.dae['g'] += [sym.im(eq_v_tc_cplx)]
    grid.dae['y_ini'] += [v_ta_r,v_tb_r,v_tc_r]
    grid.dae['y_ini'] += [v_ta_i,v_tb_i,v_tc_i]
    grid.dae['y_run'] += [v_ta_r,v_tb_r,v_tc_r]
    grid.dae['y_run'] += [v_ta_i,v_tb_i,v_tc_i]

        
    V_1 = 400/np.sqrt(3)
    V_b = V_1
    #    V_1 = 400/np.sqrt(3)*np.exp(1j*np.deg2rad(0))
    # A_1toabc = np.array([1, alpha**2, alpha])
    #V_abc = V_1 * A_1toabc 
    #e_an_r,e_bn_r,e_cn_r = V_abc.real
    #e_an_i,e_bn_i,e_cn_i = V_abc.imag

    ## inputs default values
    grid.dae['u_ini_dict'].update(
                      {f'e_ao_m_{name}':V_1,
                       f'e_bo_m_{name}':V_1,
                       f'e_co_m_{name}':V_1,
                       f'e_no_m_{name}':0.0})
    grid.dae['u_ini_dict'].update({
                       f'v_ra_{name}':0.0,
                       f'v_rb_{name}':0.0,
                       f'v_rc_{name}':0.0,
                       f'v_rn_{name}':0.0})
    grid.dae['u_run_dict'].update({
                       f'e_ao_m_{name}':V_1,
                       f'e_bo_m_{name}':V_1,
                       f'e_co_m_{name}':V_1,
                       f'e_no_m_{name}':0.0})
    grid.dae['u_run_dict'].update(
                      {f'v_ra_{name}':0.0,
                       f'v_rb_{name}':0.0,
                       f'v_rc_{name}':0.0,
                       f'v_rn_{name}':0.0})

    #u_dict.update({f'phi_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'phi_a_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'phi_b_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'phi_c_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'phi_n_{name}':0.0})

    grid.dae['u_ini_dict'].update({f'p_c_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'omega_{name}_ref':1.0})
    grid.dae['u_ini_dict'].update({f'q_ref_{name}':0.0})

    grid.dae['u_run_dict'].update({f'phi_a_{name}':0.0})
    grid.dae['u_run_dict'].update({f'phi_b_{name}':0.0})
    grid.dae['u_run_dict'].update({f'phi_c_{name}':0.0})
    grid.dae['u_run_dict'].update({f'phi_n_{name}':0.0})

    grid.dae['u_run_dict'].update({f'p_c_{name}':0.0})
    grid.dae['u_run_dict'].update({f'omega_{name}_ref':1.0})
    grid.dae['u_run_dict'].update({f'q_ref_{name}':0.0})

    grid.dae['u_ini_dict'].update({f'p_soc_{name}':0.0})
    grid.dae['u_run_dict'].update({f'p_soc_{name}':0.0})

    grid.dae['u_ini_dict'].update({f'v_dc_{name}':800.0})
    grid.dae['u_run_dict'].update({f'v_dc_{name}':800.0})

    #for ph in ['a','b','c','n']:
    #    u_dict.pop(f'i_{name}_{ph}_r')
    #    u_dict.pop(f'i_{name}_{ph}_i')

    ## parameters default values
    grid.dae['params_dict'].update({f'X_v_{name}':data['X_v'],f'R_v_{name}':data['R_v']})
    grid.dae['params_dict'].update({f'S_n_{name}':data['S_n']})
    grid.dae['params_dict'].update({f'U_n_{name}':data['U_n']})
    grid.dae['params_dict'].update({f'T_e_{name}':data['T_e']})
    grid.dae['params_dict'].update({f'T_c_{name}':data['T_c']})
    grid.dae['params_dict'].update({f'T_v_{name}':data['T_v']})
    grid.dae['params_dict'].update({f'Droop_{name}':data['Droop']})
    grid.dae['params_dict'].update({f'K_p_{name}':data['K_p']})
    grid.dae['params_dict'].update({f'T_p_{name}':data['T_p']})
    grid.dae['params_dict'].update({f'K_agc_{name}':data['K_agc']})
    grid.dae['params_dict'].update({f'K_delta_{name}':data['K_delta']})
    grid.dae['params_dict'].update({f'K_qp_{name}':data['K_qp']})
    grid.dae['params_dict'].update({f'K_qi_{name}':data['K_qi']})
    grid.dae['params_dict'].update({f'K_soc_{name}':1.0})
        

    grid.dae['u_ini_dict'].pop(str(v_ta_r))
    grid.dae['u_ini_dict'].pop(str(v_tb_r))
    grid.dae['u_ini_dict'].pop(str(v_tc_r))
    grid.dae['u_run_dict'].pop(str(v_ta_r))
    grid.dae['u_run_dict'].pop(str(v_tb_r))
    grid.dae['u_run_dict'].pop(str(v_tc_r))
    grid.dae['u_ini_dict'].pop(str(v_ta_i))
    grid.dae['u_ini_dict'].pop(str(v_tb_i))
    grid.dae['u_ini_dict'].pop(str(v_tc_i))
    grid.dae['u_run_dict'].pop(str(v_ta_i))
    grid.dae['u_run_dict'].pop(str(v_tb_i))
    grid.dae['u_run_dict'].pop(str(v_tc_i))


    
    grid.dae['xy_0_dict'].update({f'omega_{name}':1.0})


    grid.dae['h_dict'].update({f'p_pos_{name}':sym.re(s_pos),
                   f'p_neg_{name}':sym.re(s_neg),
                   f'p_zer_{name}':sym.re(s_zer)})
    grid.dae['h_dict'].update({f'q_pos_{name}':sym.im(s_pos)})
    grid.dae['h_dict'].update({str(e_ao_m):e_ao_m,str(e_bo_m):e_bo_m,str(e_co_m):e_co_m})
    grid.dae['h_dict'].update({str(v_ra):v_ra,str(v_rb):v_rb,str(v_rc):v_rc})
    grid.dae['h_dict'].update({str(p_c):p_c,str(omega_ref):omega_ref})
    grid.dae['h_dict'].update({str(phi):phi})

    HS_coi  = S_n
    omega_coi_i = S_n*omega

    grid.omega_coi_numerator += omega_coi_i 
    grid.omega_coi_denominator += HS_coi


def test_ib():
    '''
    test grid former connected to infinite bus
    '''
    import numpy as np
    import sympy as sym
    import json
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db
    import pytest

    grid = urisi('gfpizv_ib.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    model = temp.model()    
    p_c = 0.5
    q_ref = 0.1
    model.ini({'p_c_A1':p_c,'q_ref_A1':q_ref},'xy_0.json')
    
    model.report_u()
    model.report_x()
    model.report_y()
    model.report_z()

    S_n = model.get_value('S_n_A1')
    assert model.get_value('omega_A1')  == pytest.approx(1)
    assert model.get_value('omega_coi') == pytest.approx(1)
    assert model.get_value('p_vsc_A1_a') == pytest.approx(p_c*S_n/3, rel=0.05)
    assert model.get_value('p_vsc_A1_b') == pytest.approx(p_c*S_n/3, rel=0.05)
    assert model.get_value('p_vsc_A1_c') == pytest.approx(p_c*S_n/3, rel=0.05)
    assert model.get_value('p_A2') == pytest.approx(-p_c*S_n, rel=0.05)
    assert model.get_value('q_A2') == pytest.approx(-q_ref*S_n, rel=0.05)

def test_iso():
    '''
    test isolated grid former feeding a load
    '''
    import numpy as np
    import sympy as sym
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db
    import pytest

    grid = urisi('gfpizv_iso.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    model = temp.model()
    p_load_ph =  50e3
    q_load_ph =  20e3
    model.ini({'p_load_A2_a':p_load_ph,'q_load_A2_a':q_load_ph,
               'p_load_A2_b':p_load_ph,'q_load_A2_b':q_load_ph,
               'p_load_A2_c':p_load_ph,'q_load_A2_c':q_load_ph,},'xy_0.json')

    model.report_x()  
    model.report_y()
    model.report_z()

    assert model.get_value('omega_A1')  == 1.0
    assert model.get_value('omega_coi') == 1.0
    assert model.get_value('p_vsc_A1_a') == pytest.approx(p_load_ph, rel=0.05)
    assert model.get_value('p_vsc_A1_b') == pytest.approx(p_load_ph, rel=0.05)
    assert model.get_value('p_vsc_A1_c') == pytest.approx(p_load_ph, rel=0.05)
    assert model.get_value('q_vsc_A1_a') == pytest.approx(q_load_ph, rel=0.05)
    assert model.get_value('q_vsc_A1_b') == pytest.approx(q_load_ph, rel=0.05)
    assert model.get_value('q_vsc_A1_c') == pytest.approx(q_load_ph, rel=0.05)


def test_sharing():
    '''
    test isolated grid former feeding a load
    '''
    import numpy as np
    import sympy as sym
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db
    import pytest

    grid = urisi('gfpizv_sharing.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    model = temp.model()
    p_load_ph =  50e3
    q_load_ph =  20e3
    model.ini({'p_load_A2_a':p_load_ph,'q_load_A2_a':q_load_ph,
               'p_load_A2_b':p_load_ph,'q_load_A2_b':q_load_ph,
               'p_load_A2_c':p_load_ph,'q_load_A2_c':q_load_ph,},'xy_0.json')

    # model.report_x()  
    # model.report_y()
    # model.report_z()

    assert model.get_value('omega_A1')  == 1.0
    assert model.get_value('omega_A3')  == 1.0
    assert model.get_value('omega_coi') == 1.0
    assert model.get_value('p_vsc_A1_a') == pytest.approx(p_load_ph/2, rel=0.05)
    assert model.get_value('p_vsc_A1_b') == pytest.approx(p_load_ph/2, rel=0.05)
    assert model.get_value('p_vsc_A1_c') == pytest.approx(p_load_ph/2, rel=0.05)
    assert model.get_value('p_vsc_A3_a') == pytest.approx(p_load_ph/2, rel=0.05)
    assert model.get_value('p_vsc_A3_b') == pytest.approx(p_load_ph/2, rel=0.05)
    assert model.get_value('p_vsc_A3_c') == pytest.approx(p_load_ph/2, rel=0.05)


if __name__ == '__main__':

    #development()
    test_iso()


import numpy as np
import sympy as sym
import json
import hjson
import os
from pydae.urisi.genapes.genapes import add_genapes
from pydae.urisi.vscs.vscs import add_vscs
from pydae.urisi.loads.loads import add_loads
from pydae.urisi.lines.lines import lines_preprocess,add_lines,add_line_monitors
from pydae.urisi.transformers.transformers import trafos_preprocess,add_trafos,add_trafo_monitors
from pydae.urisi.shunts.shunts import add_shunts,shunts_preprocess
from pydae.urisi.sources.sources import add_sources
from pydae.urisi.ess.ess import add_ess

import pydae.build_cffi as db
from pydae.build_v2 import builder

import requests


class urisi:
    '''
    

    Parameters
    ----------
    data_input : string or dict
        File path to the system data information or dictionary with the information.

    Returns
    -------
    dict
        Dictionary with the equations for pydae. 
        
    {
     'sys':{'name':'pf_1','S_base':100e6},       
     'buses':[{'name':'GRI','P_W':0.0,'Q_var':0.0,'U_kV':66.0, 'type':'slack'},
              {'name':'POI','P_W':0.0,'Q_var':0.0,'U_kV':66.0},
              {'name':'PMV','P_W':0.0,'Q_var':0.0,'U_kV':20.0}],
     'lines':[{'bus_j':'GRI','bus_k':'POI','X_km':0.4,'R_km':0.12,'km':20},
              {'bus_j':'POI','bus_k':'PMV','X_pu':0.04,'R_pu':0.01, 'S_mva':50.0}]
    }
        

    '''

    def __init__(self,data_input=''):
        
        if type(data_input) == str:
            if 'http' in data_input:
                url = data_input
                resp = requests.get(url)
                data = json.loads(resp.text)
            else:
                if os.path.splitext(data_input)[1] == '.json':
                    with open(data_input,'r') as fobj:
                        data = json.loads(fobj.read().replace("'",'"'))
                if os.path.splitext(data_input)[1] == '.hjson':
                    with open(data_input,'r') as fobj:
                        data = hjson.loads(fobj.read().replace("'",'"'))
        elif type(data_input) == dict:
            data = data_input
            
        self.data = data
        self.N_branches = 0
        self.preprocess()

        self.dae = {'f':[],'g':[],'x':[],'y_ini':[],'y_run':[],
                    'u_ini_dict':{},'u_run_dict':{},'params_dict':{},
                    'h_dict':{},'xy_0_dict':{}}

        self.uz_jacs = True     
        self.verbose = False 

        self.omega_coi_numerator = 0.0
        self.omega_coi_denominator = 0.0

        self.A = sym.zeros(self.N_branches,self.N_nodes)
        self.At = sym.zeros(self.N_branches,self.N_nodes)

        self.G_primitive = sym.zeros(self.N_branches,self.N_branches)
        self.B_primitive = sym.zeros(self.N_branches,self.N_branches)

    def preprocess(self):

        if not 'lines' in self.data:
            self.data['lines'] = []
        if not 'shunts' in self.data:
            self.data['shunts'] = []
        if not 'transformers' in self.data:
            self.data['transformers'] = []

        self.system = self.data['system']
        self.buses = self.data['buses']
        self.lines = self.data['lines']
        self.shunts = self.data['shunts']
        self.transformers = self.data['transformers']
        
    
        self.params_grid = {'S_base':self.system['S_base']}
        self.S_base = sym.Symbol("S_base", real=True) 
        self.N_bus = len(self.buses)

        self.nodes_list = []
        self.N_nodes = 0
        for bus in self.buses:
            if 'nodes' in bus:
                bus['N_nodes'] = len(bus['nodes'])
            elif 'N_nodes' in bus:
                bus['nodes'] = list(range(bus['N_nodes']))
            else:
                bus['N_nodes'] = 4
                bus['nodes'] = list(range(bus['N_nodes']))

            self.N_nodes += bus['N_nodes']

            for node in bus['nodes']:
                self.nodes_list += [f"{bus['name']}.{node}"]

        self.it_branch = 0 # current branch
        lines_preprocess(self)
        trafos_preprocess(self)
        shunts_preprocess(self)



        #self.N_branch = 3*len(self.lines) + len(self.shunts) + 2*len(self.transformers)


    def contruct_grid(self):

        self.it_branch = 0 # current branch

        add_lines(self)
        add_trafos(self)
        add_shunts(self)

        self.G = (self.A.T @ self.G_primitive) @ self.A
        self.B = (self.A.T @ self.B_primitive) @ self.A    
        
 
        # vector of unknown voltages 
        V_list = []
        for item in self.nodes_list:
            name = item.replace('.','_')
            V_r = sym.Symbol(f'V_{name}_r', real = True)
            V_i = sym.Symbol(f'V_{name}_i', real = True)
            V_list += [V_r + sym.I*V_i]

        V = sym.Matrix(V_list)
        self.V = V

        self.I_lines = (self.G_primitive + sym.I*self.B_primitive) @ self.A @ V

        self.it_branch = 0
        add_line_monitors(self)
        add_trafo_monitors(self)

        # main complex equations 
        g_cplx = (self.G + sym.I*self.B)@V
        self.g_cplx = g_cplx
        
        # from complex equations to DAE g and y
        self.dae['g'] = []
        self.dae['y_ini']= []
        self.dae['y_run']= []
        for it in range(len(g_cplx)):
            self.dae['g'] += [sym.re(g_cplx[it])]
            self.dae['g'] += [sym.im(g_cplx[it])]

            self.dae['y_ini'] += [sym.re(V[it])]
            self.dae['y_ini'] += [sym.im(V[it])]
            self.dae['y_run'] += [sym.re(V[it])]
            self.dae['y_run'] += [sym.im(V[it])]


        # monitor voltages
        for bus in self.buses:
            if not 'monitor' in bus:
                bus.update({'monitor':False})  # do not monitor by default
            
            if bus['monitor']:
                n2a = {0:'a',1:'b',2:'c'}
                # phase top neutral voltages:
                name = bus['name']
                V_n_r = sym.Symbol(f'V_{name}_{3}_r', real = True)
                V_n_i = sym.Symbol(f'V_{name}_{3}_i', real = True)

                # phase-neutral voltage module
                for ph in [0,1,2]:
                    V_ph_r = sym.Symbol(f'V_{name}_{ph}_r', real = True)
                    V_ph_i = sym.Symbol(f'V_{name}_{ph}_i', real = True)
                    z_name = f'V_{name}_{n2a[ph]}n'
                    z_value = ((V_ph_r-V_n_r)**2 + (V_ph_i-V_n_i)**2)**0.5
                    self.dae['h_dict'].update({z_name:z_value})
                # neutral-ground voltage module
                z_name = f'V_{name}_ng'
                z_value = ((V_n_r)**2 + (V_n_i)**2)**0.5
                self.dae['h_dict'].update({z_name:z_value})

                # phase-phase voltage module
                for phj,phk in [(0,1),(1,2),(2,0)]:
                    V_phj_r = sym.Symbol(f'V_{name}_{phj}_r', real = True)
                    V_phj_i = sym.Symbol(f'V_{name}_{phj}_i', real = True)
                    V_phk_r = sym.Symbol(f'V_{name}_{phk}_r', real = True)
                    V_phk_i = sym.Symbol(f'V_{name}_{phk}_i', real = True)

                    z_name = f'V_{name}_{n2a[phj]}{n2a[phk]}'
                    z_value = ((V_phj_r-V_phk_r)**2 + (V_phj_i-V_phk_i)**2)**0.5
                    self.dae['h_dict'].update({z_name:z_value})               

    
        # voltages initial guess
        phi_deg_default = -30.0
        if 'phi_deg_default' in self.system:
            phi_deg_default = self.system['phi_deg_default']
        phi_default = np.deg2rad(phi_deg_default)
        print(phi_default)


        for bus in self.buses:
            
            V_phg = bus['U_kV']*1000.0/np.sqrt(3)
           
            if 'phi_deg_0' in bus:
                phi_0 = np.deg2rad(phi_default + bus['phi_deg_0'])
            else: phi_0 = np.deg2rad(phi_deg_default)

            for node in [0,1,2,3]:
                name = f"{bus['name']}_{node}"
                if node == 0:
                    phi = phi_0
                    V_ini = V_phg*np.exp(1j*phi)
                if node == 1:
                    phi = phi_0 - 2/3*np.pi
                    V_ini = V_phg*np.exp(1j*phi)
                if node == 2:
                    phi = phi_0 + 2/3*np.pi
                    V_ini = V_phg*np.exp(1j*phi)
                if node == 3:
                    V_ini = 0.0
        
                self.dae['xy_0_dict'].update({f'V_{name}_r':V_ini.real,f'V_{name}_i':V_ini.imag})


    def node2idx(self,bus,phase):
        '''
        Function to obtain the indexes for real and imaginary equations 
        '''
        a2n = {'a':0,'b':1,'c':2,'n':3}
        idx_r = 2*self.nodes_list.index(f'{bus}.{a2n[phase]}')
        idx_i = idx_r+1

        return idx_r,idx_i


    def construct(self, name):
        
        self.contruct_grid()          

        if 'loads' in self.data:
            add_loads(self)

        #if 'syns' in self.data:
        #    add_syns(self)
        if 'vscs' in self.data:
            add_vscs(self)
        #if 'vsgs' in self.data:
        #    add_vsgs(self)
        if 'genapes' in  self.data:
            add_genapes(self)
        #if 'wecs' in  self.data:
        #    add_wecs(self)
        if 'sources' in  self.data:
            add_sources(self)
        if 'pvs' in  self.data:
            for item in self.data['pvs']:
                add_pv(self)
        if 'ess' in  self.data:
            for item in self.data['ess']:
                add_ess(self)
                
        omega_coi = sym.Symbol("omega_coi", real=True)  

        # if self.omega_coi_denominator <1e-6:
        #     self.omega_coi_denominator = 1e-6
        #     self.omega_coi_numerator = 1e-6

        self.dae['g'] += [ -omega_coi + self.omega_coi_numerator/self.omega_coi_denominator]
        self.dae['y_ini'] += [ omega_coi]
        self.dae['y_run'] += [ omega_coi]
        self.dae['xy_0_dict'].update({'omega_coi':1.0})

        # secondary frequency control
        xi_freq = sym.Symbol("xi_freq", real=True) 
        p_agc = sym.Symbol("p_agc", real=True)  
        K_p_agc = sym.Symbol("K_p_agc", real=True) 
        K_i_agc = sym.Symbol("K_i_agc", real=True) 
        K_xif  = sym.Symbol("K_xif", real=True)
        u_freq   = sym.Symbol("u_freq", real=True)
        epsilon_freq = 1-omega_coi
        
        f_agc = [epsilon_freq - K_xif*xi_freq + u_freq]
        x_agc = [ xi_freq]

        g_agc = [ -p_agc + K_p_agc*epsilon_freq + K_i_agc*xi_freq ]
        y_agc = [  p_agc]
        

        self.dae['g'] += g_agc
        self.dae['y_ini'] += y_agc
        self.dae['y_run'] += y_agc
        self.dae['f'] += f_agc
        self.dae['x'] += x_agc
        self.dae['params_dict'].update({'K_p_agc':self.system['K_p_agc'],'K_i_agc':self.system['K_i_agc']})
        self.dae['h_dict'].update({'xi_freq':xi_freq}) 
        self.dae['u_ini_dict'].update({'u_freq':0.0}) 
        self.dae['u_run_dict'].update({'u_freq':0.0}) 
        self.dae['h_dict'].update({'u_freq':u_freq}) 
        self.dae['h_dict'].update({'u_freq':u_freq}) 
        
        if 'K_xif' in self.system:
            self.dae['params_dict'].update({'K_xif':self.system['K_xif']})
        else:
            self.dae['params_dict'].update({'K_xif':0.0})
            
        with open('xy_0.json','w') as fobj:
            fobj.write(json.dumps(self.dae['xy_0_dict'],indent=4))

        sys_dict = {'name':name,'uz_jacs':self.uz_jacs,
                'params_dict':self.dae['params_dict'],
                'f_list':self.dae['f'],
                'g_list':self.dae['g'] ,
                'x_list':self.dae['x'],
                'y_ini_list':self.dae['y_ini'],
                'y_run_list':self.dae['y_run'],
                'u_run_dict':self.dae['u_run_dict'],
                'u_ini_dict':self.dae['u_ini_dict'],
                'h_dict':self.dae['h_dict']}

        self.sys_dict = sys_dict 

    def compile(self, name):

        bldr = builder(self.sys_dict,verbose=self.verbose);
        bldr.sparse = False
        bldr.mkl = False
        bldr.uz_jacs = False
        bldr.dict2system()
        bldr.functions()
        bldr.jacobians()
        bldr.cwrite()
        bldr.template()
        bldr.compile()  

    def compile_mkl(self, name):

        b = builder(self.sys_dict,verbose=self.verbose)
        b.sparse = True
        b.mkl = True
        b.uz_jacs = self.uz_jacs
        b.dict2system()
        b.functions()
        b.jacobians()
        b.cwrite()
        b.template()
        b.compile_mkl()  

    def build(self, name=''):
        if name == '':
            print('Error: name is not provided.')
        self.construct(name)    
        self.compile(name)  


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from pydae.urisi.vscs.ac3ph4wgfpi2 import ac3ph4wgfpi2
from pydae.urisi.vscs.ac3ph4wgfpi3 import ac3ph4wgfpi3
from pydae.urisi.vscs.ac_3ph_4w_l import ac_3ph_4w_l
from pydae.urisi.vscs.acdc_3ph_4w_vdc_q import acdc_3ph_4w_vdc_q
from pydae.urisi.vscs.acdc_3ph_4w_pq import acdc_3ph_4w_pq
from pydae.urisi.vscs.dcdc_gfl import dcdc_gfl
from pydae.urisi.vscs.dcdc_gfh import dcdc_gfh
from pydae.urisi.vscs.dcdc_ph import dcdc_ph
from pydae.urisi.vscs.ac_3ph_4w_pq import ac_3ph_4w_pq
from pydae.urisi.vscs.ac_3ph_4w import ac_3ph_4w
from pydae.urisi.vscs.breaker import breaker
from pydae.urisi.vsgs.vsgs import add_vsg
from pydae.urisi.ess.ess import add_ess
from pydae.urisi.pvs.pvs import add_pv
from pydae.urisi.vsc_ctrls.vsc_ctrls import  add_vsc_ctrl


def add_vscs(grid):

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]

    for item in grid.data['vscs']:

        if 'bus' in item:
            name = item['bus']
            bus_name = item['bus']
        if 'bus_ac' in item:
            name = item['bus_ac'] 
            bus_name = item['bus_ac']           
        if 'name' in item:
            name = item['name']

        

        if item['type'] == 'ac_3ph_4w_l':
            ac_3ph_4w_l(grid,item)
        if item['type'] == 'ac3ph4wgfpi2':
            ac3ph4wgfpi2(grid,item)
        if item['type'] == 'ac3ph4wgfpi3':
            ac3ph4wgfpi3(grid,item)
        if item['type'] == 'acdc_3ph_4w_vdc_q': 
            acdc_3ph_4w_vdc_q(grid,item)
        if item['type'] == 'acdc_3ph_4w_pq': 
            acdc_3ph_4w_pq(grid,item)
        if item['type'] == 'dcdc_gfl': 
            dcdc_gfl(grid,item)
        if item['type'] == 'dcdc_gfh': 
            dcdc_gfh(grid,item)
        if item['type'] == 'dcdc_ph': 
            dcdc_ph(grid,item)
        if item['type'] == 'ac_3ph_4w_pq': 
            ac_3ph_4w_pq(grid,item)
        if item['type'] == 'ac_3ph_4w': 
            ac_3ph_4w(grid,item)
        if item['type'] == 'breaker': 
            breaker(grid,item)

        if 'vsg' in item:         
            add_vsg(grid,item['vsg'],name,bus_name)  
        if 'ess' in item:         
            add_ess(grid,item['ess'],name,bus_name)  
        if 'pv' in item:         
            add_pv(grid,item['pv'],name,bus_name)  
        if 'vsc_ctrl' in item:  
            vsc_data = item
            add_vsc_ctrl(grid,vsc_data,item['vsc_ctrl'],name,bus_name)  

    if 'vsgs' in grid.data:     
        for item in grid.data['vsgs']:    
            add_vsg(grid,item)



# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

from pydae.urisi.ess.bess_dcdc import bess_dcdc
from pydae.urisi.ess.bess_dcac import bess_dcac
from pydae.urisi.ess.bess_r import bess_r

def add_ess(grid,item,name,bus_name):

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]

    if item['type'] == 'bess_dcdc':
        bess_dcdc(grid,item,name,bus_name)
    if item['type'] == 'bess_dcac':
        bess_dcac(grid,item,name,bus_name)       
    if item['type'] == 'bess_r':
        bess_r(grid,item,name,bus_name)    


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from pydae.urisi.vsc_ctrls.ctrl_3ph_4w_pq import ctrl_3ph_4w_pq
from pydae.urisi.vsc_ctrls.ctrl_3ph_4w_droop import ctrl_3ph_4w_droop

def add_vsc_ctrl(grid,vsc_data,item,name,bus_name):

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]
    if item['type'] == 'ctrl_3ph_4w_pq':
        ctrl_3ph_4w_pq(grid,vsc_data,item,name,bus_name)            

    if item['type'] == 'ctrl_3ph_4w_droop':
        ctrl_3ph_4w_droop(grid,vsc_data,item,name,bus_name)            


import sympy as sym
import numpy as np

def pv_mpt(grid,data,name,bus_name): 
    '''

    PV+DCDC model MPPT control is implemented. 

    parameters
    ----------

    S_n: nominal power in VA
    U_n: nominal rms phase to phase voltage in V
    F_n: nominal frequency in Hz
    X_s: coupling reactance in pu (base machine S_n)
    R_s: coupling resistance in pu (base machine S_n)

    inputs
    ------

    p_s_ref: active power reference (pu, S_n base)
    q_s_ref: reactive power reference (pu, S_n base)
    v_dc: dc voltage in pu (when v_dc = 1 and m = 1, v_ac = 1)

    example
    -------

    "vscs": [{"bus":bus_name,"type":"pv_mpt_dcdc",
                 "S_n":1e6,"U_n":400.0,"F_n":50.0,
                 "X_s":0.1,"R_s":0.01,"monitor":True,
                 "I_sc":3.87,"V_oc":42.1,"I_mp":3.56,"V_mp":33.7,
                 "K_vt":-0.160,"K_it":0.065,
                 "N_pv_s":25,"N_pv_p":250}]
    
    '''
    
    ## inputs
    p_dc = sym.Symbol(f'p_dc_{name}',real=True)
    temp_deg,irrad = sym.symbols(f"temp_deg_{name},irrad_{name}", real=True)
    p_pv_ref  = sym.Symbol(f'p_pv_ref_{name}',real=True)
    p_r_ref   = sym.Symbol(f'p_r_ref_{name}',real=True)
    p_ref   = sym.Symbol(f'p_ref_{name}',real=True)

    ## parameters
    K_vt,K_it = sym.symbols(f"K_vt_{name},K_it_{name}", real=True)
    V_oc,V_mp,I_sc,I_mp = sym.symbols(f"V_oc_{name},V_mp_{name},I_sc_{name},I_mp_{name}", real=True)
    T_stc_k,i,v = sym.symbols(f"T_stc_k_{name},i_{name},v_{name}", real=True)
    v_dc,K_it = sym.symbols(f"v_dc_{name},K_it_{name}", real=True)
    N_pv_s,N_pv_p = sym.symbols(f"N_pv_s_{name},N_pv_p_{name}", real=True)
    K_mp_p,K_mp_i = sym.symbols(f"K_mp_p_{name},K_mp_i_{name}", real=True)
    T_mp,p_l_f = sym.symbols(f"T_mp_{name},p_l_f_{name}", real=True)
    p_l,xi_p = sym.symbols(f"p_l_{name},xi_p_{name}", real=True)

    T_stc_deg = 25.0

    V_oc_t = N_pv_s*V_oc * (1 + K_vt/100.0*(temp_deg - T_stc_deg))
    V_mp_t = N_pv_s*V_mp * (1 + K_vt/100.0*(temp_deg - T_stc_deg))
    I_sc_t = N_pv_p*I_sc * (1 + K_it/100.0*(temp_deg - T_stc_deg))
    I_mp_t = N_pv_p*I_mp * (1 + K_it/100.0*(temp_deg - T_stc_deg))
    I_mp_i = I_mp_t*irrad/1000.0

    v_1,i_1 = V_mp_t,I_mp_i
    v_2,i_2 = V_oc_t,0

    p_mp = V_mp_t*I_mp_i
    epsilon_p = p_mp - p_dc
    p_mppt_ref = K_mp_p*epsilon_p + K_mp_i*xi_p + p_mp
    mppt = sym.Piecewise((0.0,p_r_ref<p_mppt_ref),(1.0,p_r_ref>=p_mppt_ref))  
    
    dxi_p = epsilon_p*mppt - xi_p*(1-mppt)*1e-8  

    grid.dae['f'] += [dxi_p]
    grid.dae['x'] += [ xi_p]

    # (v_1 - v)/(v_1 - v_2) = (i_1 - i)/(i_1 - i_2)
    i_pv = p_dc/v_dc
    g_v_dc = -v_dc + v_1 - (i_1 - i_pv)*(v_1 - v_2)/(i_1 - i_2) 
    g_p_ref = - p_ref + mppt*p_mppt_ref + (1.0-mppt)*p_r_ref

    grid.dae['g'] +=     [g_v_dc, g_p_ref]
    grid.dae['y_ini'] += [v_dc, p_ref]  
    grid.dae['y_run'] += [v_dc, p_ref]  

    grid.dae['u_ini_dict'].update({f'{str(irrad)}':1000.0})
    grid.dae['u_run_dict'].update({f'{str(irrad)}':1000.0})

    grid.dae['u_ini_dict'].update({f'{str(temp_deg)}':25.0})
    grid.dae['u_run_dict'].update({f'{str(temp_deg)}':25.0})

    P_pv_n_N = data['N_pv_s']*data['V_mp']*data['N_pv_p']*data['I_mp']
    grid.dae['u_ini_dict'].update({f'{str(p_r_ref)}':1.2*P_pv_n_N})
    grid.dae['u_run_dict'].update({f'{str(p_r_ref)}':1.2*P_pv_n_N})

    grid.dae['params_dict'].update({
                   str(I_sc):data['I_sc'],
                   str(I_mp):data['I_mp'],
                   str(V_mp):data['V_mp'],
                   str(V_oc):data['V_oc'],
                   str(K_vt):data['K_vt'],
                   str(K_it):data['K_it'],
                   str(N_pv_s):data['N_pv_s'],
                   str(N_pv_p):data['N_pv_p'],
                   str(K_mp_p):data['K_mp_p'],
                   str(K_mp_i):data['K_mp_i']
                   })

    grid.dae['xy_0_dict'].update({f"v_dc_{name}":data['V_mp']*data['N_pv_s']})
    grid.dae['h_dict'].update({f"p_mp_{name}":p_mp})


def test():
    import numpy as np
    import sympy as sym
    import hjson
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db
    import pytest

    grid = urisi('pv_mpt.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    model = temp.model()

    with open('pv_mpt.hjson') as fobj:
        data = hjson.load(fobj)

    p_r_ref = 100e3
    q_ref =  50e3
    model.ini({'p_r_ref_A1':p_r_ref, 'q_ref_A1':q_ref,'N_pv_p_A1':100},'xy_0.json')
    
    assert model.get_value('p_A2') == pytest.approx(-p_r_ref, rel=0.05)
    assert model.get_value('q_A2') == pytest.approx(-q_ref, rel=0.05)

    V_mp = data['vscs'][0]['pv']['V_mp']
    I_mp = data['vscs'][0]['pv']['I_mp']
    N_pv_s = 25
    N_pv_p = 100
    p_r_ref = 500e3
    q_ref = -50e3
    p_mp = N_pv_s*V_mp*N_pv_p*I_mp
    model.ini({'p_r_ref_A1':p_r_ref, 'q_ref_A1':q_ref,'N_pv_p_A1':N_pv_p},'xy_0.json')

    assert model.get_value('p_dc_A1') == pytest.approx(p_mp, rel=0.01)
    assert model.get_value('p_A2') == pytest.approx( -p_mp, rel=0.1)
    assert model.get_value('q_A2') == pytest.approx(-q_ref, rel=0.05)


if __name__ == '__main__':

    #development()
    test()






import sympy as sym
import numpy as np
from sympy import interpolating_spline


def bess_r(grid,data,name,bus_name):
    '''

    Battery Energy Storage System modeled as a voltage source 
    behind a resistor. 

    parameters
    ----------

    S_n: nominal power in VA
    U_n: nominal rms phase to phase voltage in V
    F_n: nominal frequency in Hz
    X_s: coupling reactance in pu (base machine S_n)
    R_s: coupling resistance in pu (base machine S_n)

    ## inputs
    
    soc_ref: state of charge desired value (pu)
    p_dc: DC power (W)

    
    ## algebraic state
    v_dc: dc voltage in pu (when v_dc = 1 and m = 1, v_ac = 1)

    example
    -------

    "vscs": [{"bus":bus_name,"type":"pv_mpt_dcdc",
                 "S_n":1e6,"U_n":400.0,"F_n":50.0,
                 "X_s":0.1,"R_s":0.01,"monitor":True,
                 "I_sc":3.87,"V_oc":42.1,"I_mp":3.56,"V_mp":33.7,
                 "K_vt":-0.160,"K_it":0.065,
                 "N_pv_s":25,"N_pv_p":250}]
    
    '''

    # commons

    ## inputs
    p_dc = sym.Symbol(f"p_dc_{name}", real=True)
    k_ini = sym.Symbol(f"k_soc_ini_{name}", real=True)

    ## algebraic states
    v_dc = sym.Symbol(f"v_dc_{name}", real=True)
    p_soc = sym.Symbol(f"p_soc_{name}", real=True)

    # Battery
    E_kWh = sym.Symbol(f"E_kWh_{name}", real=True)
    soc_min = sym.Symbol(f"soc_min_{name}", real=True)
    soc_max = sym.Symbol(f"soc_max_{name}", real=True)
    K_p = sym.Symbol(f"K_soc_p_{name}", real=True)
    K_i = sym.Symbol(f"K_soc_i_{name}", real=True)
    B_0 = sym.Symbol(f"B_0_{name}", real=True)
    B_1 = sym.Symbol(f"B_1_{name}", real=True)
    R_bat = sym.Symbol(f"R_bat_{name}", real=True)
    soc_ref = sym.Symbol(f"soc_ref_{name}", real=True) 
    soc = sym.Symbol(f"soc_{name}", real=True) 
    xi_soc = sym.Symbol(f"xi_soc_{name}", real=True) 
    p_soc = sym.Symbol(f"p_soc_{name}", real=True) 

    i_dc_nosat = (p_dc)/v_dc
    i_dc = sym.Piecewise((i_dc_nosat,(i_dc_nosat>-500) & (i_dc_nosat<500)),(-500,i_dc_nosat<-500),(500,i_dc_nosat>500),(i_dc_nosat,True))
    i_dc = i_dc_nosat
    epsilon_soc = soc_ref - soc
    E_n = 1000*3600*E_kWh

    if 'socs' in data:
        socs = np.array(data['socs'])
        es = np.array(data['es'])
        e_min = np.min(es)  # minimun voltage
        e_max = np.max(es)  # maximum voltage
        interpolation = interpolating_spline(1, soc, socs, es)
        interpolation._args = tuple(list(interpolation._args) + [sym.functions.elementary.piecewise.ExprCondPair(e_max,True)])
        e = interpolation
        soc_ref_N = data['soc_0']
        e_ini = np.interp(soc_ref_N,socs,es)
    else:
        e = B_0 + B_1*soc
        e_ini = data['B_0']

    ## dynamic equations    
    dsoc = 1.0/E_n*(-i_dc*e)  # state of charge
    dxi_soc = epsilon_soc # soc controller integrator

    ## algebraic equations 
    g_v_dc = e - i_dc*R_bat - v_dc   # DC voltage
    g_p_soc = -p_soc - (K_p*epsilon_soc  + K_i*xi_soc) # soc controller power reference

    grid.dae['f'] += [dsoc, dxi_soc]
    grid.dae['x'] += [ soc,  xi_soc]

    grid.dae['g'] +=     [g_v_dc, g_p_soc]
    grid.dae['y_ini'] += [  v_dc,   p_soc]  
    grid.dae['y_run'] += [  v_dc,   p_soc]  

    soc_ref_N = 0.5
    if 'soc_0' in data:
        soc_ref_N = data['soc_0']

    grid.dae['u_ini_dict'].update({f'{str(soc_ref)}':soc_ref_N})
    grid.dae['u_run_dict'].update({f'{str(soc_ref)}':soc_ref_N})

    ## parameters  
    grid.dae['params_dict'].update({f"{K_p}":5.0}) 
    grid.dae['params_dict'].update({f"{K_i}":0.01})  
    grid.dae['params_dict'].update({f"{soc_min}":0.0}) 
    grid.dae['params_dict'].update({f"{soc_max}":1.0})           
    grid.dae['params_dict'].update({f"{E_kWh}":data['E_kWh']}) 

    if 'R_bat' in data:
        R_bat_N = data['R_bat']
    else:
        R_bat_N = 0.0

    grid.dae['params_dict'].update({f"{R_bat}":R_bat_N}) 

    grid.dae['xy_0_dict'].update({f"v_dc_{name}":e_ini})
    grid.dae['xy_0_dict'].update({f"soc_{name}":data['soc_0']})
    grid.dae['xy_0_dict'].update({f"xi_soc_{name}":0.0})
    grid.dae['xy_0_dict'].update({f"p_soc_{name}":0.0})

    if f'p_soc_{name}' in grid.dae['u_ini_dict']:
        grid.dae['u_ini_dict'].pop(f'p_soc_{name}')
        grid.dae['u_run_dict'].pop(f'p_soc_{name}')
    if f'v_dc_{name}' in grid.dae['u_ini_dict']:
        grid.dae['u_ini_dict'].pop(f'v_dc_{name}')
        grid.dae['u_run_dict'].pop(f'v_dc_{name}')


def test_vsg():
    import numpy as np
    import sympy as sym
    import json
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db

    grid = urisi('bess_r.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    S_n = 100e3
    V_n = 800
    I_n = S_n/V_n
    Conduction_losses = 0.02*S_n # = A*I_n**2
    losses = 1.0
    A = Conduction_losses/(I_n**2)*losses
    B = 1*losses
    C = 0.02*S_n*losses
    model = temp.model()
    model.Dt = 1.0
    params = {'A_loss_A1':A,'B_loss_A1':B,'C_loss_A1':C,'R_bat_A1':0.001,'p_c_A1':0.0,
            'K_soc_p_A1':0.5,'K_soc_i_A1':0.0001}
    model.ini(params,'xy_0.json')
    # model.save_xy_0('xy_1.json')
    # model.report_params()
    # model.report_u()
    model.report_x()
    model.report_y()
    model.report_z()

def test_pq():
    import numpy as np
    import sympy as sym
    import json
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db

    grid = urisi('bess_r_ctrl_3ph_4w_pq.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    S_n = 100e3
    V_n = 800
    I_n = S_n/V_n
    Conduction_losses = 0.02*S_n # = A*I_n**2
    losses = 1.0
    A = Conduction_losses/(I_n**2)*losses
    B = 1*losses
    C = 0.02*S_n*losses
    model = temp.model()
    model.Dt = 1.0
    params = {'A_loss_A1':A,'B_loss_A1':B,'C_loss_A1':C,'R_bat_A1':0.001,'q_ref_A1':50e3, 
            'K_soc_p_A1':0.5,'K_soc_i_A1':0.0001}
    model.ini(params,'xy_0.json')
    # model.save_xy_0('xy_1.json')
    # model.report_params()
    # model.report_u()
    model.report_x()
    model.report_y()
    model.report_z()

if __name__ == '__main__':

    #development()
    test_pq()






# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym


def tgov1(dae,syn_data,name,bus_name):
    '''
    Governor TGOV1 like in PSS/E
    
    '''
    bus_name = syn_data['bus']
    gov_data = syn_data['gov']
    
    # inpunts
    omega = sym.Symbol(f"omega_{name}", real=True)
    p_agc = sym.Symbol(f"p_agc", real=True)
    p_c = sym.Symbol(f"p_c_{name}", real=True) 

    # dynamic states
    x_gov_1 = sym.Symbol(f"x_gov_1_{name}", real=True)
    x_gov_2 = sym.Symbol(f"x_gov_2_{name}", real=True)  

    # algebraic states
    p_m = sym.Symbol(f"p_m_{name}", real=True)  
    p_m_ref = sym.Symbol(f"p_m_ref_{name}", real=True)  

    # parameters
    T_1 = sym.Symbol(f"T_gov_1_{name}", real=True)  # 1
    T_2 = sym.Symbol(f"T_gov_2_{name}", real=True)  # 2
    T_3 = sym.Symbol(f"T_gov_3_{name}", real=True)  # 10
    D_t = sym.Symbol(f"D_t_{name}", real=True)  # 10 
    Droop = sym.Symbol(f"Droop_{name}", real=True)  # 0.05
    K_sec = sym.Symbol(f"K_sec_{name}", real=True)  # 0.05
   
    omega_ref = sym.Symbol(f"omega_ref_{name}", real=True)

    # auxiliar
    Domega = (omega - omega_ref)
    p_r = K_sec*p_agc

    # differential equations
    dx_gov_1 =   (p_m_ref - x_gov_1)/T_1
    dx_gov_2 =   (x_gov_1 - x_gov_2)/T_3

    g_p_m_ref  = -p_m_ref + p_c + p_r - 1/Droop*Domega
    g_p_m = (x_gov_1 - x_gov_2)*T_2/T_3 + x_gov_2 - D_t*Domega - p_m

    
    dae['f'] += [dx_gov_1,dx_gov_2]
    dae['x'] += [ x_gov_1, x_gov_2]
    dae['g'] += [g_p_m_ref,g_p_m]
    dae['y_ini'] += [  p_m_ref,  p_m]  
    dae['y_run'] += [  p_m_ref,  p_m]  

    dae['params_dict'].update({str(Droop):gov_data['Droop']})
    dae['params_dict'].update({str(T_1):gov_data['T_1']})
    dae['params_dict'].update({str(T_2):gov_data['T_2']})
    dae['params_dict'].update({str(T_3):gov_data['T_3']})
    dae['params_dict'].update({str(D_t):gov_data['D_t']})
    dae['params_dict'].update({str(K_sec):gov_data['K_sec']})

    dae['params_dict'].update({str(omega_ref):1.0})

    dae['u_ini_dict'].update({str(p_c):gov_data['p_c']})
    dae['u_run_dict'].update({str(p_c):gov_data['p_c']})

    dae['h_dict'].update({str(p_c):gov_data['p_c']})


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def vsource(grid,name,bus_name,data_dict):
    '''

    parameters
    ----------



    inputs
    ------

    theta_ref: internal voltage angle reference
    v_ref: internal voltage magnitude reference

    example
    -------

    "vsource": [{"v_ref":1.0, "theta_ref":0.0}]

    
    '''

    sin = sym.sin
    cos = sym.cos

    
    # inputs
    V = sym.Symbol(f"V_{bus_name}", real=True)
    theta = sym.Symbol(f"theta_{bus_name}", real=True)
    v_ref = sym.Symbol(f"v_ref_{name}", real=True)
    theta_ref = sym.Symbol(f"theta_ref_{name}", real=True)
    V_dummy = sym.Symbol(f"V_dummy_{name}", real=True)

    
    # dynamic states

    # algebraic states
    # V = sym.Symbol(f"V_{name}", real=True)
    # theta = sym.Symbol(f"theta_{name}", real=True)            


    # parameters
    params_list = []
    
    # auxiliar

    # dynamic equations            
    grid.dae['f'] += [V_dummy - v_ref]
    grid.dae['x'] += [V_dummy]

    # algebraic equations   
    g_V = V - v_ref
    g_theta = theta - theta_ref
    
    
    # dae 

    H = 1e15
    grid.H_total += H
    grid.omega_coi_numerator += H
    grid.omega_coi_denominator += H

    idx_V = grid.dae['y_ini'].index(V)
    idx_theta = grid.dae['y_ini'].index(theta)

    grid.dae['g'][idx_V] = g_V
    grid.dae['g'][idx_theta] = g_theta

    # grid.dae['y_ini'] += [V, theta]  
    # grid.dae['y_run'] += [V, theta]  

    grid.dae['u_ini_dict'].update({f'{str(v_ref)}':1.0})
    grid.dae['u_run_dict'].update({f'{str(v_ref)}':1.0})

    grid.dae['u_ini_dict'].update({f'{str(theta_ref)}':0.0})
    grid.dae['u_run_dict'].update({f'{str(theta_ref)}':0.0})

    grid.dae['xy_0_dict'].update({str(V):1.0})
    grid.dae['xy_0_dict'].update({str(theta):0.0})
       
    grid.dae['h_dict'].update({f"V_dummy_{name}":V_dummy})

    # outputs


def test_mkl():

    import numpy as np
    from pydae.bmapu import bmapu_builder
    from pydae.build_v2 import builder
    from pydae.utils import read_data
    import json

    grid = bmapu_builder.bmapu('vsource.hjson')
    grid.construct(f'temp')
    grid.compile_mkl('temp')

    import temp

    model = temp.model()
    model.ini({},'xy_0.json')
    model.report_y()


def test():

    from pydae.bmapu import bmapu_builder
    
    grid = bmapu_builder.bmapu('vsource.hjson')
    grid.checker()
    grid.verbose = True 
    grid.build('temp')

    import temp

    model = temp.model()
    model.ini({},'xy_0.json')
    model.report_y()


 
if __name__=='__main__':
    test_mkl()








# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from pydae.urisi.sources.vdc_src import vdc_src
from pydae.urisi.sources.ac3ph4w_ideal import ac3ph4w_ideal
from pydae.urisi.sources.ac3ph3w_ideal import ac3ph3w_ideal
def add_sources(grid):

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]

    for item in grid.data['sources']:
        if item['type'] == 'vdc_src':
            vdc_src(grid,item)
        if item['type'] == 'ac3ph4w_ideal':
            ac3ph4w_ideal(grid,item)
        if item['type'] == 'ac3ph3w_ideal':
            ac3ph3w_ideal(grid,item)


import numpy as np
import matplotlib.pyplot as plt
import ipywidgets as widgets
import pydae.svg_tools as st 
import pydae.build_cffi as db
from pydae.bmapu import bmapu_builder

plt.style.use('https://raw.githubusercontent.com/pydae/pydae/master/src/pydae/edashboards/presentation.mplstyle')
colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

class dashboard():
    
    def __init__(self):

        
        self.colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
        self.params = {'v_dc_ref_1':1.35,"K_pdc_1":100,'irrad_1':1000,'mode_1':2}

        self.data_pv = {"I_sc":3.87,"V_oc":42.1,"I_mpp":3.56,"V_mpp":33.7,"N_s":72,
        "K_vt":-0.160,"K_it":0.065,"R_pv_s": 0.5602, "R_pv_sh": 1862, "K_d": 1.3433}

    def ini(self):

        import pv_mid

        self.model = pv_mid.model()
        self.model.Dt = 0.01
        self.model.decimation = 1
        self.model.ini(self.params,'xy_0.json')
        
        
    def build(self):

        data = {
        "system":{"name":"smib","S_base":100e6, "K_p_agc":0.0,"K_i_agc":0.0,"K_xif":0.01},       
        "buses":[{"name":"1", "P_W":0.0,"Q_var":0.0,"U_kV":20.0},
                    {"name":"2", "P_W":0.0,"Q_var":0.0,"U_kV":20.0}
                ],
        "lines":[{"bus_j":"1", "bus_k":"2", "X_pu":0.05,"R_pu":0.0,"Bs_pu":0.0,"S_mva":100.0}],
        "pvs":[
            {"type":"pv_1","bus":"1","S_n":1e6,"U_n":400.0,      
             "R_s":0.01,"X_s":0.05,
             "K_pdc":100,"C_dc":10.5,
             "I_sc":8,"V_oc":42.1,"I_mpp":3.56,"V_mpp":33.7,"N_s":72,
             "K_vt":-0.160,"K_it":0.065,"R_pv_s": 0.5602, "R_pv_sh": 1862, "K_d": 1.3433,
             "N_ms":25,"N_mp":250}],
        "sources":[{"type":"genape", "bus":"2","S_n":100e6,"F_n":50.0,"X_v":0.001,"R_v":0.0,"K_delta":0.001,"K_alpha":1e-6}]
        }

        data['pvs'][0].update(self.data_pv)

        grid = bmapu_builder.bmapu(data)
        #grid.checker()
        grid.uz_jacs = True
        grid.verbose = False
        grid.build('pv_mid')

    def widgets(self):
        
        model = self.model
        colors = self.colors

        plt.ioff()
        plt.clf()

        self.s = st.svg(r"pv_dashboard.svg")

        self.html = widgets.HTML(
            value= self.s.tostring(),
            placeholder='',
            description='',
        )

        tab_0 = widgets.VBox([widgets.Text(description='Sn')])
        self.tab_1_sld_irrad = widgets.FloatSlider(description='irrad',min=0, max=1200, step=100, value=1000)
        self.tab_1_sld_temp_deg = widgets.FloatSlider(description='temp (deg)',min=10, max=75, step=1, value=25)
        tab_1 = widgets.HBox([self.tab_1_sld_irrad,self.tab_1_sld_temp_deg])

        tab_2_0 = widgets.RadioButtons(
            options=['Manual', 'Power from MPPT', 'Power from speed control'],
            description='Mode:',    disabled=False)
        self.sld_p_s  = widgets.FloatSlider(description='p<sub>s</sub>*',min=0, max=1, step=0.1, value=0)
        self.sld_v_dc = widgets.FloatSlider(description='v<sub>dc</sub>*',min=0.8, max=2.0, step=0.1, value=1.2)
        self.sld_q_s  = widgets.FloatSlider(description='q<sub>s</sub>*',min=-1, max=1, step=0.1, value=0)
        self.sld_i_d  = widgets.FloatSlider(description='i<sub>d</sub>*',min=-1, max=1, step=0.1, value=0)
        self.sld_i_q  = widgets.FloatSlider(description='i<sub>q</sub>*',min=-0.1, max=0.1, step=0.1, value=0)

        tab_2_1 = widgets.VBox([self.sld_p_s,
                                ])
        tab_2 = widgets.HBox([tab_2_0,tab_2_1])

        self.options_mode = widgets.RadioButtons(
            options=['Normal', 'LVRT'],
            description='Mode:',    disabled=False)

        tab_3_0 = widgets.HBox([self.options_mode,widgets.VBox([self.sld_v_dc,self.sld_q_s]),widgets.VBox([self.sld_i_d,self.sld_i_q])])
        tab_3 = widgets.HBox([tab_3_0])

        self.sld_V_g = widgets.FloatSlider(description='V<sub>g</sub>',min=0.2, max=1.2, step=0.1, value=1.0)
        self.sld_SCR = widgets.FloatSlider(description='SCR',min=10, max=100, step=1, value=10)

        tab_grid = widgets.HBox([self.sld_V_g,self.sld_SCR])

        self.tab = widgets.Tab()
        self.tab.children = [tab_1,tab_3,tab_grid]

        #tab.set_title(0, 'PV')
        self.tab.set_title(0, 'Enviroment')
        #tab.set_title(2, 'MPPT')
        self.tab.set_title(1, 'VSC')
        self.tab.set_title(2, 'Grid')

        self.params.update({'v_dc_ref_1':self.sld_v_dc.value,
                            "K_pdc_1":100,'irrad_1':1000,'b_1_2':-0.1,'g_1_2':0.0})

        model.ini(self.params,'xy_0.json')
        
    def update(self,change):
        
        model = self.model
        S_n_1_MVA = model.get_value('S_n_1')/1e6
        S_b_s_MVA = 100.0
        b_1_2 = -self.sld_SCR.value*S_n_1_MVA/S_b_s_MVA
        self.params.update({'v_dc_ref_1':self.sld_v_dc.value,"K_pdc_1":100,'q_s_ref_1':self.sld_q_s.value,
                'irrad_1':self.tab_1_sld_irrad.value,'temp_deg_1':self.tab_1_sld_temp_deg.value,
                'i_sd_i_ref_1':self.sld_i_d.value, 'i_sq_i_ref_1':self.sld_i_q.value, 
                'b_1_2':b_1_2,
                'v_ref_2':self.sld_V_g.value})
        model.ini(self.params,'xy_0.json')
        
        if self.options_mode.value == 'LVRT':
            model.ini({'v_dc_ref_1':1.8,"K_pdc_1":1,'q_s_ref_1':0.0,
                    'irrad_1':self.tab_1_sld_irrad.value,'temp_deg_1':self.tab_1_sld_temp_deg.value,
                    'i_sd_i_ref_1':self.sld_i_d.value, 'i_sq_i_ref_1':self.sld_i_q.value, 
                    'b_1_2':b_1_2,
                    'v_ref_2':self.sld_V_g.value})

        # s.set_tspan('p_g', f'={model.get_value("p_g_"):0.2f}')
        # s.set_tspan('q_g', f'={model.get_value("q_g_"):0.2f}')
        I_s = (model.get_value("i_si_1")**2 + model.get_value("i_sr_1")**2)**0.5
        p_s = model.get_value("p_s_1")
        q_s = model.get_value("q_s_1")   
        s_s = p_s+ 1j*q_s
        phi_s = np.angle(s_s) 
        pf_s = np.sign(q_s)*np.cos(phi_s)

        self.s.set_tspan('I_s', f'={I_s:4.2f}')
        self.s.set_tspan('i_d_ref', f'={model.get_value("i_sd_ref_1"):5.2f}')
        self.s.set_tspan('i_q_ref', f'={model.get_value("i_sq_ref_1"):5.2f}')
        self.s.set_tspan('m', f'={model.get_value("m_ref_1"):0.2f}')
        self.s.set_tspan('p_s', f'={p_s:5.2f}')
        self.s.set_tspan('q_s', f'={q_s:5.2f}')
        self.s.set_tspan('v_dc', f'={model.get_value("v_dc_1"):0.2f}')
        self.s.set_tspan('i_dc', f'={model.get_value("i_pv_pu_1"):0.2f}')
        self.s.set_tspan('v_pv', f'={model.get_value("v_pv_1"):4.1f} V')
        self.s.set_tspan('i_pv', f'={model.get_value("i_pv_1"):4.1f} A')
        self.s.set_tspan('p_pv', f'={model.get_value("p_pv_1"):5.1f} W')
        self.s.set_tspan('V_pcc', f'={model.get_value("V_1"):5.2f}')
        self.s.set_tspan('V_g', f'={model.get_value("V_2"):5.2f}')
        self.s.set_tspan('pf_s', f'={pf_s:5.2f}')

        # s.set_tspan('beta', f'={np.abs(model.get_value("beta_")):5.2f}')

        self.html.value = self.s.tostring() 


    def show(self):

        # Link the slider value to the text element update function
        self.tab_1_sld_irrad.observe(self.update, names='value')
        self.tab_1_sld_temp_deg.observe(self.update, names='value')
        self.sld_v_dc.observe(self.update, names='value')
        self.sld_q_s.observe(self.update, names='value')
        self.sld_i_d.observe(self.update, names='value')
        self.sld_i_q.observe(self.update, names='value')
        self.sld_V_g.observe(self.update, names='value')
        self.sld_SCR.observe(self.update, names='value')
        self.options_mode.observe(self.update, names='value')

        self.update(0)
        self.layout = widgets.VBox([self.html,self.tab])

        display(self.layout)


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def leon_vsg_ll(grid,name,bus_name,data_dict):
    """
    # auxiliar equations
    Omega_b = 2*np.pi*F_n
    omega_s = omega_coi
    v_D = V*sin(theta)  # e^(-j)
    v_Q = V*cos(theta) 
    v_d = v_D * cos(delta) - v_Q * sin(delta)   
    v_q = v_D * sin(delta) + v_Q * cos(delta)

    Domega = x_v + K_p * (p_ref - p)
    e_dv = 0.0
    epsilon_v = v_ref - V
    i_d = i_d_ref
    i_q = i_q_ref
    omega_v = Domega + 1.0
    q_ref_0 = K_p_v * epsilon_v + K_i_v * xi_v

    # dynamical equations
    ddelta   = Omega_b*(omega_v - omega_s) - K_delta*delta
    dx_v = K_i*(p_ref - p) - K_g*(omega_v - 1.0)
    de_qm = 1.0/T_q * (q - q_ref_0 - q_ref - e_qm) 
    dxi_v = epsilon_v # PI agregado

    # algebraic equations
    g_i_d_ref  = e_dv - R_v * i_d_ref - X_v * i_q_ref - v_d 
    g_i_q_ref  = e_qv - R_v * i_q_ref + X_v * i_d_ref - v_q 
    g_p  = v_d*i_d + v_q*i_q - p  
    g_q  = v_d*i_q - v_q*i_d - q 
    g_e_qv = 1.0 - e_qv - K_q*e_qm 

    {"bus":"1","type":"vsg_ll",'S_n':10e6,'F_n':50,'K_delta':0.0,
    'R_v':0.01,'X_v':0.1,'K_p':1.0,'K_i':0.1,'K_g':0.0,'K_q':20.0,
    'T_q':0.1,'K_p_v':1e-6,'K_i_v':1e-6}
    
    """

    sin = sym.sin
    cos = sym.cos

    ctrl_data = data_dict['ctrl']

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]                  

    idx_bus = buses_list.index(bus_name) # get the number of the bus where the syn is connected
    if not 'idx_powers' in buses[idx_bus]: buses[idx_bus].update({'idx_powers':0})
    buses[idx_bus]['idx_powers'] += 1

    # inputs:
    q_l,q_r,v_ref = sym.symbols(f'q_l_{name},q_r_{name},v_ref_{name}', real=True)
    p_l,p_r = sym.symbols(f'p_l_{name},p_r_{name}', real=True)
    v_dc = sym.Symbol(f"v_dc_{name}", real=True)   
    Ddelta_ff,Domega_ff = sym.symbols(f'Ddelta_ff_{name},Domega_ff_{name}', real=True)

    # dynamical states:
    delta,x_v,e_qm,xi_v = sym.symbols(f'delta_{name},x_v_{name},e_qm_{name},xi_v_{name}', real=True)
    
    # algebraic states:
    p_s = sym.Symbol(f'p_s_{name}', real=True)
    q_s = sym.Symbol(f'q_s_{name}', real=True)
    e_vq = sym.Symbol(f"e_vq_{name}", real=True)
    v_td_ref = sym.Symbol(f"v_td_ref_{name}", real=True)
    v_tq_ref = sym.Symbol(f"v_tq_ref_{name}", real=True) 

    # params:
    S_base = sym.Symbol('S_base', real = True) # S_base = global power base, S_n = machine power base
    S_n,F_n,K_delta = sym.symbols(f'S_n_{name},F_n_{name},K_delta_{name}', real=True)
    K_p,K_i,K_g,K_i_q = sym.symbols(f'K_p_{name},K_i_{name},K_g_{name},K_i_q_{name}', real=True)
    R_v,X_v = sym.symbols(f'R_v_{name},X_v_{name}', real=True)
    K_q,T_q = sym.symbols(f'K_q_{name},T_q_{name}', real=True)
    K_p_v,K_i_v = sym.symbols(f'K_p_v_{name},K_i_v_{name}', real=True)
    params_list = ['F_n','K_delta','K_p','K_i','K_g','R_v','X_v','K_q','T_q','K_p_v','K_i_v']

    # auxiliar variables and constants
    omega_coi = sym.Symbol("omega_coi", real=True) # from global system
    V_s = sym.Symbol(f"V_{bus_name}", real=True) # from global system
    theta_s = sym.Symbol(f"theta_{bus_name}", real=True) # from global system
    i_si = sym.Symbol(f"i_si_{name}", real=True)
    i_sr = sym.Symbol(f"i_sr_{name}", real=True)  
    m = sym.Symbol(f"m_{name}", real=True) 
    theta_t = sym.Symbol(f"theta_t_{name}", real=True) 

    # auxiliar equations
    Omega_b = 2*np.pi*F_n
    omega_s = omega_coi

    i_sD = i_si  # e^(-j)
    i_sQ = i_sr


    delta_ff =  delta + Ddelta_ff
    i_sd = i_sD * cos(delta_ff) - i_sQ * sin(delta_ff)   
    i_sq = i_sD * sin(delta_ff) + i_sQ * cos(delta_ff)

    p_ref = p_l + p_r
    q_ref = q_l + q_r
    Domega = x_v + K_p * (p_ref - p_s)
    e_vd = 0.0
    epsilon_v = v_ref - V_s
    omega_v = Domega + 1.0 + Domega_ff
    q_ref_0 = K_p_v * epsilon_v + K_i_v * xi_v 

    v_tD_ref =  v_td_ref * cos(delta_ff) + v_tq_ref * sin(delta_ff)   
    v_tQ_ref = -v_td_ref * sin(delta_ff) + v_tq_ref * cos(delta_ff) 

    v_ti_ref = v_tD_ref
    v_tr_ref = v_tQ_ref   
    m_ref = sym.sqrt(v_tr_ref**2 + v_ti_ref**2)/v_dc
    theta_t_ref = sym.atan2(v_ti_ref,v_tr_ref) 
    

    # dynamical equations
    ddelta   = Omega_b*(omega_v - omega_s) - K_delta*delta
    dx_v = K_i*(p_ref - p_s) - K_g*(omega_v - 1.0)
    de_qm = 1.0/T_q * (q_s - q_ref_0 - q_ref - e_qm) 
    dxi_v = epsilon_v # PI agregado

    # algebraic equations
    g_v_td_ref  = e_vd - R_v * i_sd - X_v * i_sq - v_td_ref 
    g_v_tq_ref  = e_vq - R_v * i_sq + X_v * i_sd - v_tq_ref 

    g_e_vq = 1.0 - e_vq - K_q*e_qm 
    g_m  = m-m_ref
    g_theta_t  = theta_t-theta_t_ref
    
    # DAE system update
    grid.dae['f'] += [ddelta,dx_v,de_qm,dxi_v]
    grid.dae['x'] += [ delta, x_v, e_qm, xi_v]
    grid.dae['g'] +=     [g_v_td_ref,g_v_tq_ref,g_e_vq,g_m,g_theta_t]
    grid.dae['y_ini'] += [  v_td_ref,  v_tq_ref,  e_vq,  m,  theta_t]
    grid.dae['y_run'] += [  v_td_ref,  v_tq_ref,  e_vq,  m,  theta_t]
            
    # default inputs
    grid.dae['u_ini_dict'].update({f'p_l_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'q_l_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'p_r_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'q_r_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'v_ref_{name}':1.0})
    grid.dae['u_ini_dict'].update({f'Ddelta_ff_{name}':0.0})
    grid.dae['u_ini_dict'].update({f'Domega_ff_{name}':0.0})

    grid.dae['u_run_dict'].update({f'p_l_{name}':0.0})
    grid.dae['u_run_dict'].update({f'q_l_{name}':0.0})
    grid.dae['u_run_dict'].update({f'p_r_{name}':0.0})
    grid.dae['u_run_dict'].update({f'q_r_{name}':0.0})
    grid.dae['u_run_dict'].update({f'v_ref_{name}':1.0})
    grid.dae['u_run_dict'].update({f'Ddelta_ff_{name}':0.0})
    grid.dae['u_run_dict'].update({f'Domega_ff_{name}':0.0})

    # default parameters
    for item in params_list:
        grid.dae['params_dict'].update({item + f'_{name}':ctrl_data[item]})

    # add speed*H term to global for COI speed computing
    H = 4.0
    grid.H_total += H
    grid.omega_coi_numerator += omega_v*H*S_n
    grid.omega_coi_denominator += H*S_n

    # DAE outputs update
    grid.dae['h_dict'].update({f"omega_v_{name}":omega_v})
    grid.dae['h_dict'].update({f"p_ref_{name}":p_ref})
    grid.dae['h_dict'].update({f"q_ref_{name}":q_ref})
    grid.dae['h_dict'].update({f"v_ref_{name}":v_ref})
    grid.dae['h_dict'].update({f"i_sd_{name}":i_sd})
    grid.dae['h_dict'].update({f"i_sq_{name}":i_sq})
    grid.dae['h_dict'].update({f"delta_ff_{name}":delta_ff})
    
    grid.dae['xy_0_dict'].update({str(e_vq):1.0}) 
    grid.dae['xy_0_dict'].update({str(v_tq_ref):1.0}) 

    p_W   = p_s * S_n
    q_var = q_s * S_n

    return p_W,q_var


if __name__ == "__main__":
    pass  


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from pydae.bmapu.avrs.sexs import sexs
from pydae.bmapu.avrs.sexs import sexsq
from pydae.bmapu.avrs.kundur import kundur
from pydae.bmapu.avrs.ntsst4 import ntsst4
from pydae.bmapu.avrs.ntsst1 import ntsst1
from pydae.bmapu.avrs.kundur_tgr import kundur_tgr

def add_avr(dae,syn_data,name,bus_name):
    
    if syn_data['avr']['type'] == 'sexs':
        sexs(dae,syn_data,name,bus_name)

    if syn_data['avr']['type'] == 'sexsq':
        sexsq(dae,syn_data,name)

    if syn_data['avr']['type'] == 'kundur':
        kundur(dae,syn_data,name,bus_name)

    if syn_data['avr']['type'] == 'avr_kundurq':
        avr_kundurq(dae,syn_data,name,bus_name)

    if syn_data['avr']['type'] == 'ntsst4':
        ntsst4(dae,syn_data,name,bus_name)

    if syn_data['avr']['type'] == 'ntsst1':
        ntsst1(dae,syn_data,name,bus_name)

    if syn_data['avr']['type'] == 'kundur_tgr':
        kundur_tgr(dae,syn_data,name,bus_name)
        



# tests/test_dynamic_module.py

import os
import importlib.util
import pytest

@pytest.fixture
def create_module():
    module_content = """
def dynamic_function(x, y):
    return x * y
"""
    module_path = os.path.join(os.path.dirname(__file__), "dynamic_module.py")

    with open(module_path, "w") as module_file:
        module_file.write(module_content)

    yield module_path

    # Clean up after the test
    #os.remove(module_path)

def test_create_module(create_module):
    assert os.path.isfile(create_module)

def test_dynamic_function(create_module):
    # Import the module dynamically
    module_name = "dynamic_module"
    spec = importlib.util.spec_from_file_location(module_name, create_module)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    # Call the dynamically created function
    result = module.dynamic_function(2, 3)
    assert result == 6



# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def olives_vsg(grid,name,bus_name,data_dict):
    """
 
    """

    sin = sym.sin
    cos = sym.cos
    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]
                
    bus_name = data_dict['bus']
    
    if 'name' in data_dict:
        name = data_dict['name']
    else:
        name = bus_name
        
    for gen_id in range(100):
        if name not in grid.generators_id_list:
            grid.generators_id_list += [name]
            break
        else:
            name = name + f'_{gen_id}'
            
    data_dict['name'] = name
                        
    idx_bus = buses_list.index(bus_name) # get the number of the bus where the syn is connected
    if not 'idx_powers' in buses[idx_bus]: buses[idx_bus].update({'idx_powers':0})
    buses[idx_bus]['idx_powers'] += 1
    
    # inputs
    V = sym.Symbol(f"V_{bus_name}", real=True)
    theta = sym.Symbol(f"theta_{bus_name}", real=True)
    p_m = sym.Symbol(f"p_m_{name}", real=True)
    v_ref = sym.Symbol(f"v_ref_{name}", real=True)  
    omega_coi = sym.Symbol("omega_coi", real=True)   
    p_c = sym.Symbol(f"p_c_{name}", real=True)
    p_agc = sym.Symbol("p_agc", real=True)
    omega_ref = sym.Symbol(f"omega_ref_{name}", real=True)
    q_ref = sym.Symbol(f"q_ref_{name}", real=True) 
        
    # dynamic states
    delta = sym.Symbol(f"delta_{name}", real=True)
    e_qv = sym.Symbol(f"e_qv_{name}", real=True)
    xi_p = sym.Symbol(f"xi_p_{name}", real=True)
    p_ef = sym.Symbol(f"p_ef_{name}", real=True)
    p_cf = sym.Symbol(f"p_cf_{name}", real=True)

    # algebraic states
    omega = sym.Symbol(f"omega_{name}", real=True)
    i_d = sym.Symbol(f"i_d_{name}", real=True)
    i_q = sym.Symbol(f"i_q_{name}", real=True)            
    p_g = sym.Symbol(f"p_g_{name}", real=True)
    q_g = sym.Symbol(f"q_g_{name}", real=True)


    # parameters
    S_n = sym.Symbol(f"S_n_{name}", real=True)
    Omega_b = sym.Symbol(f"Omega_b_{name}", real=True)            
    K_p = sym.Symbol(f"K_p_{name}", real=True)    
    K_q = sym.Symbol(f"K_q_{name}", real=True)   
    T_p = sym.Symbol(f"T_p_{name}", real=True)
    X_v = sym.Symbol(f"X_v_{name}", real=True)
    T_v = sym.Symbol(f"T_v_{name}", real=True)
    X_v = sym.Symbol(f"X_v_{name}", real=True)
    R_v = sym.Symbol(f"R_v_{name}", real=True)
    K_delta = sym.Symbol(f"K_delta_{name}", real=True)
    Droop = sym.Symbol(f"Droop_{name}", real=True) 
    K_sec = sym.Symbol(f"K_sec_{name}", real=True) 
    T_e  = sym.Symbol(f"T_e_{name}", real=True)  
    T_c  = sym.Symbol(f"T_c_{name}", real=True)  

    params_list = ['S_n','Omega_b','K_p','T_p','K_q','T_v','X_v','R_v','K_delta','K_sec','Droop','K_sec']
    
    # auxiliar
    v_d = V*sin(delta - theta) 
    v_q = V*cos(delta - theta) 
    p_e = i_d*(v_d + R_v*i_d) + i_q*(v_q + R_v*i_q)     
    omega_s = omega_coi
    e_dv = 0
    p_r = K_sec*p_agc
    epsilon_p = p_m - p_ef
    epsilon_q = q_ref - q_g
    
                        
    
    # dynamic equations            
    ddelta = Omega_b*(omega - omega_s) - K_delta*delta
    dxi_p = epsilon_p
    de_qv = 1/T_v*(v_ref + K_q*epsilon_q - e_qv)
    dp_ef = 1/T_e*(p_e - p_ef)
    dp_cf = 1/T_c*(p_c - p_cf)

    # algebraic equations   
    g_omega = -omega + K_p*(epsilon_p + xi_p/T_p) + 1
    g_i_d  = -R_v*i_d + X_v*i_q - v_d + e_dv 
    g_i_q  = -R_v*i_q - X_v*i_d - v_q + e_qv
    g_p_g  = i_d*v_d + i_q*v_q - p_g  
    g_q_g  = i_d*v_q - i_q*v_d - q_g 
    g_p_m  = -p_m + p_cf + p_r - 1/Droop*(omega - omega_ref)
    
    # dae 
    f_vsg = [ddelta,dxi_p,de_qv,dp_ef,dp_cf]
    x_vsg = [ delta, xi_p, e_qv, p_ef, p_cf]
    g_vsg = [g_omega,g_i_d,g_i_q,g_p_g,g_q_g,g_p_m]
    y_vsg = [  omega,  i_d,  i_q,  p_g,  q_g,  p_m]
    
    # T_p = K_p*2*H
    H = T_p/(2*K_p)
    grid.H_total += H
    grid.omega_coi_numerator += omega*H*S_n
    grid.omega_coi_denominator += H*S_n

    grid.dae['f'] += f_vsg
    grid.dae['x'] += x_vsg
    grid.dae['g'] += g_vsg
    grid.dae['y_ini'] += y_vsg  
    grid.dae['y_run'] += y_vsg  
    
    if 'v_ref' in data_dict:
        grid.dae['u_ini_dict'].update({f'{str(v_ref)}':data_dict['v_ref']})
        grid.dae['u_run_dict'].update({f'{str(v_ref)}':data_dict['v_ref']})
    else:
        grid.dae['u_ini_dict'].update({f'{str(v_ref)}':1.0})
        grid.dae['u_run_dict'].update({f'{str(v_ref)}':1.0})

    #if 'p_m' in data_dict:
    #    grid.dae['u_ini_dict'].update({f'{str(p_m)}':data_dict['p_m']})
    #    grid.dae['u_run_dict'].update({f'{str(p_m)}':data_dict['p_m']})
    #else:
    #    grid.dae['u_ini_dict'].update({f'{str(p_m)}':1.0})
    #    grid.dae['u_run_dict'].update({f'{str(p_m)}':1.0})
        
    grid.dae['u_ini_dict'].update({f'{str(p_c)}':data_dict['p_c']})
    grid.dae['u_run_dict'].update({f'{str(p_c)}':data_dict['p_c']})

    grid.dae['u_ini_dict'].update({f'{str(omega_ref)}':1.0})
    grid.dae['u_run_dict'].update({f'{str(omega_ref)}':1.0})

    grid.dae['u_ini_dict'].update({f'{str(q_ref)}':0.0})
    grid.dae['u_run_dict'].update({f'{str(q_ref)}':0.0})
    

    grid.dae['xy_0_dict'].update({str(omega):1.0})
    
    # outputs
    grid.dae['h_dict'].update({f"p_e_{name}":p_e})
    
    # 
    for item in params_list:       
        grid.dae['params_dict'].update({f"{item}_{name}":data_dict[item]})

    if 'T_e' in data_dict:
        grid.dae['params_dict'].update({f'{str(T_e)}':data_dict['T_e']})
    else:
        grid.dae['params_dict'].update({f'{str(T_e)}':0.1})

    if 'T_c' in data_dict:
        grid.dae['params_dict'].update({f'{str(T_c)}':data_dict['T_c']})
    else:
        grid.dae['params_dict'].update({f'{str(T_c)}':0.1})


    grid.dae['params_dict'].update({f"{item}_{name}":data_dict[item]})

    p_W   = p_g * S_n
    q_var = q_g * S_n

    return p_W,q_var




# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""


import sympy as sym


def ntsst4(dae,syn_data,name,bus_name):
    '''

    .. table:: Constants
        :widths: auto

        ================== =========== ============================================= =========== 
        Variable           Code        Description                                   Units
        ================== =========== ============================================= ===========
        :math:`T_A`        ``T_a``     Time Constant                                 s
        ================== =========== ============================================= ===========

    Example:
    
    ``"avr":{"type":"ntsst4","K_pr":3.15,"K_ir":3.15,"V_rmax":1.0,"V_rmin":-0.87,"T_a":0.02,"K_pm":1.0,"K_im":0.0,"K_p": 6.5,"v_ref":1.0},``

    '''

    avr_data = syn_data['avr']
    remote_bus_name = bus_name
    if 'bus' in avr_data:
        remote_bus_name = avr_data['bus']
    
    v_t = sym.Symbol(f"V_{remote_bus_name}", real=True)   
    v_c = sym.Symbol(f"v_c_{remote_bus_name}", real=True)  
    xi_v  = sym.Symbol(f"xi_v_{name}", real=True)
    x_a   = sym.Symbol(f"x_a_{name}", real=True)
    xi_m  = sym.Symbol(f"xi_m_{name}", real=True)
    v_f = sym.Symbol(f"v_f_{name}", real=True)  
    K_pr   = sym.Symbol(f'K_pr_{name}', real=True) 
    K_ir   = sym.Symbol(f'K_ir_{name}', real=True) 
    T_a    = sym.Symbol(f'T_a_{name}', real=True) 
    K_pm   = sym.Symbol(f'K_pm_{name}', real=True) 
    K_im   = sym.Symbol(f'K_im_{name}', real=True) 
    K_p    = sym.Symbol(f'K_p_{name}', real=True)  

    #V_rmax = sym.Symbol('V_rmax ', real=True) 
    #V_rmin = sym.Symbol('V_rmin ', real=True) 
    #V_mmax = sym.Symbol('V_mmax ', real=True) 
    #V_mmin = sym.Symbol('V_mmin ', real=True) 
    #K_g    = sym.Symbol('K_g    ', real=True) 
    #K_i    = sym.Symbol('K_i    ', real=True) 
    #T_r    = sym.Symbol('T_r    ', real=True) 
    #V_bmax = sym.Symbol('V_bmax ', real=True) 
    #K_c    = sym.Symbol('K_c    ', real=True) 
    #X_l    = sym.Symbol('X_l    ', real=True) 
    #theta_p= sym.Symbol('theta_p', real=True) 
    
    v_ref = sym.Symbol(f"v_ref_{name}", real=True) 
    v_pss = sym.Symbol(f"v_pss_{name}", real=True) 

    v_s = v_pss # v_oel and v_uel are not considered
    #v_ini = K_ai*xi_v
    v_c = v_t

    epsilon_v = v_ref - v_c + v_s
    v_r = K_pr*epsilon_v + K_ir*xi_v
    v_e = K_p*v_c

    epsilon_m = x_a # -K_g*v_f;
    v_m = K_pm*epsilon_m + K_im*xi_m
 
    dxi_v = epsilon_v
    dx_a = 1.0/T_a*(v_r - x_a)
    dxi_m = epsilon_m
      
    g_v_f  =  v_m*v_e - v_f 
    
    dae['f'] += [dxi_v,dx_a,dxi_m]
    dae['x'] += [ xi_v, x_a, xi_m]
    dae['g'] += [g_v_f]
    dae['y_ini'] += [v_f] 
    dae['y_run'] += [v_f]  

    dae['params_dict'].update({str(K_pr):avr_data['K_pr']}) 
    dae['params_dict'].update({str(K_ir):avr_data['K_ir']}) 
    dae['params_dict'].update({str(T_a):avr_data['T_a']})  
    dae['params_dict'].update({str(K_pm):avr_data['K_pm']}) 
    dae['params_dict'].update({str(K_im):avr_data['K_im']}) 
    dae['params_dict'].update({str(K_p):avr_data['K_p']})  

    dae['u_ini_dict'].update({str(v_ref):avr_data['v_ref']})
    dae['u_run_dict'].update({str(v_ref):avr_data['v_ref']})

    dae['u_run_dict'].update({str(v_pss):0.0})
    dae['u_ini_dict'].update({str(v_pss):0.0})

    dae['xy_0_dict'].update({str(xi_v):1.0})
    dae['xy_0_dict'].update({str(x_a):1.0})
    dae['xy_0_dict'].update({str(xi_m):1.0})
    
def test():
    import numpy as np
    import sympy as sym
    import hjson
    from pydae.bmapu.bmapu_builder import bmapu
    import pydae.build_cffi as db
    import pytest

    grid = bmapu('ntsst4.hjson')
    grid.checker()
    grid.uz_jacs = True
    grid.build('temp')

    import temp

    model = temp.model()

    v_ref_1 = 1.0
    model.ini({},'xy_0.json')

    model.report_x()
    model.report_y()

    assert model.get_value('V_2') == pytest.approx(v_ref_1, rel=0.001)


if __name__ == '__main__':

    #development()
    test()



def ac3ph3wpq(grid,vsc_data):
    '''
    Converter type p_ac,q_ac 3 phase 3 wire
    
    '''
    
    ## Model data
    bus_ac_name = vsc_data['bus_ac']
    bus_dc_name = vsc_data['bus_dc']  
    a_value  = vsc_data['a']   
    b_value  = vsc_data['b']   
    c_value  = vsc_data['c']   
    
    ## Voltages:
    v_a_r,v_a_i = sym.symbols(f'v_{bus_ac_name}_a_r,v_{bus_ac_name}_a_i',real=True)
    v_b_r,v_b_i = sym.symbols(f'v_{bus_ac_name}_b_r,v_{bus_ac_name}_b_i',real=True)
    v_c_r,v_c_i = sym.symbols(f'v_{bus_ac_name}_c_r,v_{bus_ac_name}_c_i',real=True)
    v_n_r,v_n_i = sym.symbols(f'v_{bus_ac_name}_n_r,v_{bus_ac_name}_n_i',real=True)
    v_dc_a_r,v_dc_n_r = sym.symbols(f'v_{bus_dc_name}_a_r,v_{bus_dc_name}_n_r',real=True)

    ## Currents:
    i_a_r,i_a_i = sym.symbols(f'i_vsc_{bus_ac_name}_a_r,i_vsc_{bus_ac_name}_a_i',real=True)
    i_b_r,i_b_i = sym.symbols(f'i_vsc_{bus_ac_name}_b_r,i_vsc_{bus_ac_name}_b_i',real=True)
    i_c_r,i_c_i = sym.symbols(f'i_vsc_{bus_ac_name}_c_r,i_vsc_{bus_ac_name}_c_i',real=True) 
    i_dc_a_r,i_dc_n_r = sym.symbols(f'i_vsc_{bus_dc_name}_a_r,i_vsc_{bus_dc_name}_n_r',real=True)


    
    ## Body:
        
    ### AC-side
    p_ac,q_ac,p_dc,p_loss = sym.symbols(f'p_vsc_{bus_ac_name},q_vsc_{bus_ac_name},p_vsc_{bus_dc_name},p_vsc_loss_{bus_ac_name}',real=True)
    coef_a,coef_b,coef_c = sym.symbols(f'coef_a_{bus_ac_name},coef_b_{bus_ac_name},coef_c_{bus_ac_name}',real=True)
    
    v_a = v_a_r + 1j*v_a_i
    v_b = v_b_r + 1j*v_b_i
    v_c = v_c_r + 1j*v_c_i
    v_n = v_n_r + 1j*v_n_i

    i_a = i_a_r + 1j*i_a_i
    i_b = i_b_r + 1j*i_b_i
    i_c = i_c_r + 1j*i_c_i

    s_a = (v_a - v_n) * sym.conjugate(i_a)
    s_b = (v_b - v_n) * sym.conjugate(i_b)
    s_c = (v_c - v_n) * sym.conjugate(i_c)

    eq_i_a_r =  sym.re(s_a) - p_ac*coef_a
    eq_i_b_r =  sym.re(s_b) - p_ac*coef_b
    eq_i_c_r =  sym.re(s_c) - p_ac*coef_c
    eq_i_a_i =  sym.im(s_a) - q_ac*coef_a
    eq_i_b_i =  sym.im(s_b) - q_ac*coef_b
    eq_i_c_i =  sym.im(s_c) - q_ac*coef_c


    i_abc_list  = [i_a_r,i_a_i,i_b_r,i_b_i,i_c_r,i_c_i]
    for itg in [1,2,3]:
        bus_idx = grid.nodes.index(f'{bus_ac_name}.{itg}')
        g_idx = bus_idx - grid.N_nodes_v
        grid.dae['g'][2*g_idx+0] += i_abc_list[2*(itg-1)  ]
        grid.dae['g'][2*g_idx+1] += i_abc_list[2*(itg-1)+1]

    ### DC side
    a,b,c = sym.symbols(f'a_{bus_ac_name},b_{bus_ac_name},c_{bus_ac_name}',real=True)
    i_rms = sym.sqrt(i_a_r**2+i_a_i**2+0.1) 
    p_simple = a + b*i_rms + c*i_rms*i_rms

    p_vsc_loss = p_simple

    eq_p_loss = p_loss - p_vsc_loss
    eq_i_dc_a_r = i_dc_a_r + p_dc/(v_dc_a_r-v_dc_n_r+1e-8)
    eq_i_dc_n_r = i_dc_n_r + p_dc/(v_dc_n_r-v_dc_a_r+1e-8)
    eq_p_dc = p_dc - p_ac - sym.Piecewise((-p_loss, p_dc < 0), (p_loss, p_dc > 0),(p_loss, True))


    ## DAE system update
    grid.dae['g'] += [eq_i_a_r,eq_i_a_i,
                      eq_i_b_r,eq_i_b_i,
                      eq_i_c_r,eq_i_c_i]
    grid.dae['y'] += [   i_a_r,   i_a_i,
                         i_b_r,   i_b_i,
                         i_c_r,   i_c_i]

    grid.dae['g'] +=  [eq_i_dc_a_r,eq_i_dc_n_r,eq_p_dc,eq_p_loss]
    grid.dae['y'] +=  [   i_dc_a_r,   i_dc_n_r,   p_dc,   p_loss]  
    grid.dae['u'].update({f'p_vsc_{bus_ac_name}':0.0,f'q_vsc_{bus_ac_name}':0.0}) 
    grid.dae['xy_0_dict'].update({f'v_{bus_dc_name}_a_r':800.0,f'v_{bus_dc_name}_n_r':10.0})
    
    #grid.dae['u'].pop(f'i_{bus_dc_name}_a_r')
    #grid.dae['u'].pop(f'i_{bus_dc_name}_n_r')
    grid.dae['params'].update({f'a_{bus_ac_name}':a_value,f'b_{bus_ac_name}':b_value,f'c_{bus_ac_name}':c_value})
    grid.dae['params'].update({f'coef_a_{bus_ac_name}':1/3,f'coef_b_{bus_ac_name}':1/3,f'coef_c_{bus_ac_name}':1/3})

    ## Add current injections to grid equations:
    bus_idx = grid.nodes.index(f'{bus_dc_name}.{1}')
    g_idx = bus_idx - grid.N_nodes_v
    grid.dae['g'][2*g_idx+0] += i_dc_a_r
    grid.dae['g'][2*g_idx+1] += 0.0   

    bus_idx = grid.nodes.index(f'{bus_dc_name}.{4}')
    g_idx = bus_idx - grid.N_nodes_v
    grid.dae['g'][2*g_idx+0] += i_dc_n_r
    grid.dae['g'][2*g_idx+1] += 0.0  
    



import sympy as sym
import numpy as np

def pv_mpt_dcdc(grid,data,name,bus_name):
    '''

    PV+DCDC model MPPT control is implemented. 

    parameters
    ----------

    S_n: nominal power in VA
    U_n: nominal rms phase to phase voltage in V
    F_n: nominal frequency in Hz
    X_s: coupling reactance in pu (base machine S_n)
    R_s: coupling resistance in pu (base machine S_n)

    inputs
    ------

    p_s_ref: active power reference (pu, S_n base)
    q_s_ref: reactive power reference (pu, S_n base)
    v_dc: dc voltage in pu (when v_dc = 1 and m = 1, v_ac = 1)

    example
    -------

    "vscs": [{"bus":bus_name,"type":"pv_mpt_dcdc",
                 "S_n":1e6,"U_n":400.0,"F_n":50.0,
                 "X_s":0.1,"R_s":0.01,"monitor":True,
                 "I_sc":3.87,"V_oc":42.1,"I_mp":3.56,"V_mp":33.7,
                 "K_vt":-0.160,"K_it":0.065,
                 "N_pv_s":25,"N_pv_p":250}]
    
    '''
    
    bus_name = data['bus']
    name = bus_name
   
    A_value  = data['A']   
    B_value  = data['B']   
    C_value  = data['C']  

    #### voltages:
    v_hp, v_hn = sym.symbols(f'V_{bus_name}_0_r,V_{bus_name}_1_r', real=True)
    v_hp_i, v_hn_i = sym.symbols(f'V_{bus_name}_0_i,V_{bus_name}_1_i', real=True)
   
    ### LV-side
    A,B,C = sym.symbols(f'A_{name},B_{name},C_{name}',real=True)

    ### HV-side
    e_h,v_h,i_h,p_h = sym.symbols(f'e_h_{name},v_h_{name},i_h_{name},p_h_{name}',real=True)
    R_h = sym.Symbol(f'R_h_{name}',real=True)
    p_r_ref = sym.Symbol(f'p_r_ref_{name}',real=True)
    m_h = sym.Symbol(f'm_h_{name}',real=True)
    i_g = sym.Symbol(f'i_g_{name}',real=True) 
    R_g = sym.Symbol(f'R_g_{name}',real=True)


    ## PV
    K_vt,K_it = sym.symbols(f"K_vt_{name},K_it_{name}", real=True)
    V_oc,V_mp,I_sc,I_mp = sym.symbols(f"V_oc_{name},V_mp_{name},I_sc_{name},I_mp_{name}", real=True)
    temp_deg,irrad = sym.symbols(f"temp_deg_{name},irrad_{name}", real=True)
    T_stc_k,i,v = sym.symbols(f"T_stc_k_{name},i_{name},v_{name}", real=True)
    v_dc,K_it = sym.symbols(f"v_dc_{name},K_it_{name}", real=True)
    N_pv_s,N_pv_p = sym.symbols(f"N_pv_s_{name},N_pv_p_{name}", real=True)
    K_pp,K_pi = sym.symbols(f"K_pp_{name},K_pi_{name}", real=True)
    T_mp,p_l_f = sym.symbols(f"T_mp_{name},p_l_f_{name}", real=True)
    p_l,xi_p = sym.symbols(f"p_l_{name},xi_p_{name}", real=True)

    T_stc_deg = 25.0

    V_oc_t = N_pv_s*V_oc * (1 + K_vt/100.0*(temp_deg - T_stc_deg))
    V_mp_t = N_pv_s*V_mp * (1 + K_vt/100.0*(temp_deg - T_stc_deg))
    I_sc_t = N_pv_p*I_sc * (1 + K_it/100.0*(temp_deg - T_stc_deg))
    I_mp_t = N_pv_p*I_mp * (1 + K_it/100.0*(temp_deg - T_stc_deg))
    I_mp_i = I_mp_t*irrad/1000.0

    v_1,i_1 = V_mp_t,I_mp_i
    v_2,i_2 = V_oc_t,0

    # (v_1 - v)/(v_1 - v_2) = (i_1 - i)/(i_1 - i_2)
    i_pv = p_l/v_dc
    p_mp = V_mp_t*I_mp_i
    #v_dc_v = v_1 - (i_1 - i_pv)*(v_1 - v_2)/(i_1 - i_2) 
    g_v_dc = -v_dc + v_1 - (i_1 - i_pv)*(v_1 - v_2)/(i_1 - i_2) 

    grid.dae['g'] +=     [g_v_dc]
    grid.dae['y_ini'] += [v_dc]  
    grid.dae['y_run'] += [v_dc]  

    grid.dae['u_ini_dict'].update({f'{str(irrad)}':1000.0})
    grid.dae['u_run_dict'].update({f'{str(irrad)}':1000.0})

    grid.dae['u_ini_dict'].update({f'{str(temp_deg)}':25.0})
    grid.dae['u_run_dict'].update({f'{str(temp_deg)}':25.0})

    grid.dae['params_dict'].update({
                   str(I_sc):data['I_sc'],
                   str(I_mp):data['I_mp'],
                   str(V_mp):data['V_mp'],
                   str(V_oc):data['V_oc'],
                   str(N_pv_s):data['N_pv_s'],
                   str(N_pv_p):data['N_pv_p'],
                   str(K_vt):data['K_vt'],
                   str(K_it):data['K_it']
                   })

    grid.dae['xy_0_dict'].update({f"v_dc_{name}":data['V_mp']*data['N_pv_s']})

    # VSC
    v_l = v_dc
    epsilon_p = p_mp - p_l 
    p_mppt_h_ref = K_pp*epsilon_p + K_pi*xi_p + p_mp
    mppt = sym.Piecewise((0.0,p_r_ref<p_mppt_h_ref),(1.0,p_r_ref>=p_mppt_h_ref))  
    p_h = mppt*p_mppt_h_ref + (1.0-mppt)*p_r_ref
    v_h = v_hp - v_hn
    i_h = p_h/v_h
    i_hp =  i_h
    i_hn = -i_h
    p_loss = A*i_hp**2 + B*sym.Abs(i_hp) + C

    eq_p_l = -p_l + p_h + p_loss

    dxi_p = epsilon_p*mppt - xi_p*(1-mppt)*1e-8  
    dp_l_f = 1/T_mp*(p_l - p_l_f)

    grid.dae['f'] += [dxi_p, dp_l_f]
    grid.dae['x'] += [ xi_p,  p_l_f]
    grid.dae['g'] += [eq_p_l]
    grid.dae['y_ini'] += [ p_l]  
    grid.dae['y_run'] += [ p_l]  

    grid.dae['u_ini_dict'].update({f'{str(p_r_ref)}':1e8})
    grid.dae['u_run_dict'].update({f'{str(p_r_ref)}':1e8})

    # current injections dc LV side
    idx_hp_r,idx_hp_i = grid.node2idx(f'{bus_name}','a')
    idx_hn_r,idx_hn_i = grid.node2idx(f'{bus_name}','b')
    grid.dae['g'] [idx_hp_r] += -i_hp 
    grid.dae['g'] [idx_hn_r] += -i_hn
    grid.dae['g'] [idx_hp_i] +=  v_hp_i/1e3
    grid.dae['g'] [idx_hn_i] +=  v_hn_i/1e3
   
    grid.dae['params_dict'].update({f'R_h_{name}':0.01})    
    grid.dae['params_dict'].update({f'R_g_{name}':3})   
    grid.dae['params_dict'].update({f'A_{name}':A_value})   
    grid.dae['params_dict'].update({f'B_{name}':B_value})   
    grid.dae['params_dict'].update({f'C_{name}':C_value})   
    grid.dae['params_dict'].update({f'K_pp_{name}':data['K_pp']})   
    grid.dae['params_dict'].update({f'K_pi_{name}':data['K_pi']})   
    grid.dae['params_dict'].update({f'T_mp_{name}':1.0})   

    grid.dae['h_dict'].update({f'p_loss_{name}':p_loss})
    grid.dae['h_dict'].update({f'i_pv_{name}':i_pv})
    grid.dae['h_dict'].update({f'p_mp_{name}':p_mp})
    grid.dae['h_dict'].update({f'p_h_{name}':p_h})

def test():
    import numpy as np
    import sympy as sym
    import json
    from pydae.urisi.urisi_builder import urisi
    import pydae.build_cffi as db

    grid = urisi('pv_mpt_dcdc.hjson')
    grid.uz_jacs = True
    grid.construct('temp')
    grid.compile('temp')

    import temp

    S_n = 500e3
    V_n = 800
    I_n = S_n/V_n
    Conduction_losses = 0.02*S_n # = A*I_n**2
    A = Conduction_losses/(I_n**2)
    B = 1*0
    C = 0.02*S_n
    model = temp.model()
    model.ini({'A_D1':A,'B_D1':B,'C_D1':C,'p_r_ref_D1':500e3},'xy_0.json')
    model.report_u()
    model.report_y()
    model.report_z()



if __name__ == '__main__':

    #development()
    test()






# -*- coding: utf-8 -*-
"""
Created on Sat Sep  5 18:14:08 2020

@author: jmmau
"""

from setuptools import setup
from Cython.Build import cythonize
import numpy

#setup(
#    ext_modules = cythonize("convolve2.pyx",include_path = [numpy.get_include()])
#)



from distutils.core import setup, Extension


# setup(
#     ext_modules=[
#         Extension("convolve2", ["convolve2.c"],
#                   include_dirs=[numpy.get_include()]),
#     ],
# )

# Or, if you use cythonize() to make the ext_modules list,
# include_dirs can be passed to setup()

setup(
    ext_modules=cythonize("convolve2.pyx"),
    include_dirs=[numpy.get_include()]
)


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym
from pydae.urisi.vsgs.vsg_lpf import vsg_lpf
from pydae.urisi.vsgs.gfpizv import gfpizv
from pydae.urisi.vsgs.gflpfzv import gflpfzv

def add_vsg(grid,item,name,bus_name):

    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]
    if item['type'] == 'vsg_lpf':
        vsg_lpf(grid,item,name,bus_name)
    if item['type'] == 'gfpizv':
        gfpizv(grid,item,name,bus_name)
    if item['type'] == 'gflpfzv':
        gflpfzv(grid,item,name,bus_name)

            


# -*- coding: utf-8 -*-
"""
Created on Thu Feb  6 07:22:05 2020

@author: Juan Manuel Mauricio
"""

import numpy as np
import numba


@numba.njit(cache=True) 
def perturbations(t,struct): 
    
    return
        
    

        
#@numba.njit(cache=True) 
def daesolver(run_eval,struct,perturbations): 
    sin = np.sin
    cos = np.cos
    sqrt = np.sqrt
    i = 0 
    
    Dt = struct[i].Dt 
    N_steps = struct[i].N_steps 
    N_store = struct[i].N_store 
    N_x = struct[i].N_x
    N_y = struct[i].N_y
    N_outs = 1 
    decimation = struct[i].decimation 
    eye = np.eye(N_x)
    # initialization 
    #t = struct[i].t 
    t = 0.0
    run_eval(0.0,struct, 1) 
    it_store = 0 
    struct[i]['T'][0] = t 
    struct[i].X[0,:] = struct[i].x[:,0]  
    
    Y = np.zeros((N_store,N_y))

    

    Y[0,:] = struct[i].y[:,0]  
 
    solver = struct[i].solvern 
    for it in range(N_steps-1): 
        t += Dt 
 
        perturbations(t,struct) 
        
        if solver == 1: 
            # forward euler solver  
            run_eval(t,struct, 2)  
            struct[i].x[:] += Dt*struct[i].f  
 
        if solver == 2: 
            
            # bacward euler solver
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run_eval(t,struct, 2) 
                run_eval(t,struct, 3) 
                run_eval(t,struct, 10)  
                phi =  x_0 + Dt*struct[i].f - struct[i].x 
                Dx = np.linalg.solve(-(Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                if np.max(np.abs(Dx)) < struct[i].itol: break 
            print(struct[i].f)
 
        if solver == 3: 
            # trapezoidal solver
            run_eval(t,struct, 2) 
            f_0 = np.copy(struct[i].f[:]) 
            x_0 = np.copy(struct[i].x[:]) 
            for j in range(struct[i].imax): 
                run_eval(t,struct, 10)  
                phi =  x_0 + 0.5*Dt*(f_0 + struct[i].f) - struct[i].x 
                Dx = np.linalg.solve(-(0.5*Dt*struct[i].Fx - np.eye(N_x)), phi) 
                struct[i].x[:] += Dx[:] 
                run_eval(t,struct, 2) 
                if np.max(np.abs(Dx)) < struct[i].itol: break 

        if solver == 4: 

            run_eval(t,struct, 2) 
            run_eval(t,struct, 3) 

            x = np.copy(struct[i].x[:]) 
            y = np.copy(struct[i].y[:]) 
            f = np.copy(struct[i].f[:]) 
            g = np.copy(struct[i].g[:]) 
            
            for iter in range(struct[i].imax):
                run_eval(t,struct, 2) 
                run_eval(t,struct, 3) 
                run_eval(t,struct,10) 
                run_eval(t,struct,11) 
                
                x_i = struct[i].x[:] 
                y_i = struct[i].y[:]  
                f_i = struct[i].f[:] 
                g_i = struct[i].g[:]                 
                F_x_i = struct[i].Fx[:,:]
                F_y_i = struct[i].Fy[:,:] 
                G_x_i = struct[i].Gx[:,:] 
                G_y_i = struct[i].Gy[:,:]                

                A_c_i = np.vstack((np.hstack((eye-0.5*Dt*F_x_i, -0.5*Dt*F_y_i)),
                                   np.hstack((G_x_i,         G_y_i))))
                     
                f_n_i = x_i - x - 0.5*Dt*(f_i+f) 
                #print(t,iter,g_i)
                Dxy_i = np.linalg.solve(-A_c_i,np.vstack((f_n_i,g_i))) 
                
                x_i = x_i + Dxy_i[0:N_x]
                y_i = y_i + Dxy_i[N_x:(N_x+N_y)]

                struct[i].x[:] = x_i
                struct[i].y[:] = y_i

                if np.max(np.abs(Dxy_i[:,0]))<struct[i].itol:
                    
                    break
                
                if iter>struct[i].imax-2:
                    
                    print('Convergence problem', iter)

            struct[i].x[:] = x_i
            struct[i].y[:] = y_i
                
        # channels 
        if it >= it_store*decimation and it_store < N_store-1: 
            struct[i]['T'][it_store+1] = t 
            struct[i].X[it_store+1,:] = struct[i].x[:,0] 
            Y[it_store+1,:] = struct[i].y[:,0]
            it_store += 1 
    struct[i].t = t
    struct[i].it_store = it_store
    return struct[i]['T'][:], struct[i].X[:], Y 


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def add_fault(grid,data):
    """
    # auxiliar
    
    """

    bus_name = data['bus']
    name = bus_name

    fault_b = sym.Symbol(f'fault_b_{name}', real=True)
    fault_g = sym.Symbol(f'fault_g_{name}', real=True)
    fault_g_ref= sym.Symbol(f'fault_g_ref_{name}', real=True)
    RampDown = sym.Symbol(f'RampDown_{name}', real=True)
    RampUp = sym.Symbol(f'RampUp_{name}', real=True)
    K_fault = sym.Symbol(f'K_fault_{name}', real=True)

    V = sym.Symbol(f'V_{name}', real=True)


    epsilon_g = fault_g_ref - fault_g
    dfault_g_nosat = K_fault*(fault_g_ref - fault_g)
    dfault_g_sat = sym.Piecewise((RampDown,dfault_g_nosat<RampDown),
                                (RampUp,dfault_g_nosat>RampUp),
                                (dfault_g_nosat,True))
    grid.dae['f'] += [dfault_g_sat]
    grid.dae['x'] += [fault_g]

    p_pu = fault_g*V**2
    q_pu = fault_b*V**2

    grid.dae['u_ini_dict'].update({f'{fault_b}':0.0})
    grid.dae['u_run_dict'].update({f'{fault_b}':0.0})
    grid.dae['u_ini_dict'].update({f'{fault_g_ref}':0.0})
    grid.dae['u_run_dict'].update({f'{fault_g_ref}':0.0})

    grid.dae['params_dict'].update({f'{str(RampDown)}':-20000})
    grid.dae['params_dict'].update({f'{str(RampUp)}':10000})
    grid.dae['params_dict'].update({f'{str(K_fault)}':10000})

    # grid power injection
    buses = grid.data['buses']
    buses_list = [bus['name'] for bus in buses]

    idx_bus = buses_list.index(bus_name) # get the number of the bus where the syn is connected
    if not 'idx_powers' in buses[idx_bus]: buses[idx_bus].update({'idx_powers':0})
    buses[idx_bus]['idx_powers'] += 1

    S_base = sym.Symbol('S_base', real = True)
    grid.dae['g'][idx_bus*2]   += -p_pu
    grid.dae['g'][idx_bus*2+1] += -q_pu


def test():
    import numpy as np
    import sympy as sym
    import hjson
    from pydae.bmapu.bmapu_builder import bmapu
    import pydae.build_cffi as db
    import pytest
    import matplotlib.pyplot as plt

    grid = bmapu('fault.hjson')
    grid.checker()
    grid.uz_jacs = True
    grid.build('temp')


    import temp

    model = temp.model()
    model.Dt = 0.001
    model.decimation = 1

    v_ref_1 = 1.0
    model.ini({'p_m_1':0.5,'v_ref_1':v_ref_1,'K_a_1':200,
            'RampDown_1':-45e3,'RampUp_1':30e3, 'K_fault_1':1e2},'xy_0.json')

    # assert model.get_value('V_1') == pytest.approx(v_ref_1, rel=0.001)

    model.run(1.0,{})
    model.run(1.1,{'fault_g_ref_1':20000})
    model.run(1.4,{})
    model.run(5,{'fault_g_ref_1':0})

    # model.run(1.11,{'fault_g_ref_1':0})
    # model.run(2,{'fault_g_ref_1':0})

    model.post()
    # # assert model.get_value('q_A2') == pytest.approx(-q_ref, rel=0.05)

    # model.ini({'p_m_1':0.5,'v_ref_1':1.0},'xy_0.json')
    # model.run(1.0,{})
    # model.run(15.0,{'v_ref_1':1.05})
    # model.post()


    fig,axes = plt.subplots()
    axes.plot(model.Time,model.get_values('omega_1'))
    #axes.plot(model.Time,model.Y)
    axes.grid()
    fig.savefig('fault.svg')

if __name__ == '__main__':

    #development()
    test()


# -*- coding: utf-8 -*-
"""
Created on Thu August 10 23:52:55 2022

@author: jmmauricio
"""

import numpy as np
import sympy as sym

def pv_1(grid,name,bus_name,data_dict):
    '''

    parameters
    ----------

    S_n: nominal power in VA
    F_n: nominal frequency in Hz
    X_v: coupling reactance in pu (base machine S_n)
    R_v: coupling resistance in pu (base machine S_n)
    K_delta: if K_delta>0.0 current generator is converted to reference machine 
    K_alpha: alpha gain to obtain Domega integral 

    inputs
    ------

    alpha: RoCoF in pu if K_alpha = 1.0
    omega_ref: frequency in pu
    v_ref: internal voltage reference

    example
    -------

    "vscs": [{"type":"vsc_l","S_n":1e6,"F_n":50.0,"X_s":0.1,"R_s":0.01}]

    v_t = m*v_dc/sqrt(6)

    v_t_pu*v_b_ac = m*v_dc_pu*v_b_dc/sqrt(6)

    S_b = S_n
    U_b = U_n
    Omega_b = 2*np.pi*F_b
    Omega_rb = Omega_rn
    Tau_b = S_b/Omega_rb
    V_dc_b = sqrt(2)*U_b
    I_b = (np.sqrt(3)*U_b)
    I_dq_b = I_b*np.sqrt(2)
    V_dq_b = U_b*np.sqrt(2/3)

    V_dq_b and I_dq_b defined as below verify that: 
    3/2*I_dq_b*V_dq_b = 3/2*I_b*np.sqrt(2)*U_b*np.sqrt(2/3) = np.sqrt(3)*U_b*I_b

    Z_b = U_b**2/S_b
    Z_b = (V_dq_b/np.sqrt(2/3))**2/S_b = 3/2*V_dq_b**2/S_b
    L_m_b = Z_b/Omega_rb

    tau_r = 3/2*Phi*N_pp*i_mq

    tau_r_pu = (3/2*Phi*N_pp*i_mq_pu)*I_dq_b/Tau_b  = K_tau * i_mq_pu
    with K_tau = 3/2*Phi*N_pp*I_dq_b/Tau_b

    0 =               - R_s*i_md - omega_r*L_m*i_mq - v_md  
    0 = omega_r*Phi_m - R_s*i_mq + omega_r*L_m*i_md - v_mq 

    0 =               (- R_s*i_md_pu/V_dq_b - omega_r*L_m*i_mq_pu/V_dq_b - v_md_pu/I_dq_b)*I_dq_b*V_dq_b
    0 =               (- R_s_pu*i_md_pu/V_dq_b - omega_r*L_m_pu*i_mq_pu*Z_b/Omega_rb/(V_dq_b*Z_b) - v_md_pu/(I_dq_b*Z_b))*I_dq_b*V_dq_b*Z_b
    0 =               (- R_s_pu*i_md_pu/V_dq_b - omega_r_pu*L_m_pu*i_mq_pu/(V_dq_b) - v_md_pu/(I_dq_b*Z_b))*I_dq_b*V_dq_b*Z_b


    v_t_pu*v_b_ac = m*V_dc_b*sqrt(2)*U_b/sqrt(6)
    v_t_pu = m*V_dc_b*sqrt(2)/sqrt(6) = m*V_dc_b*sqrt(3)


    

    '''

    sin = sym.sin
    cos = sym.cos

    # inputs
    V_s = sym.Symbol(f"V_{bus_name}", real=True)
    theta_s = sym.Symbol(f"theta_{bus_name}", real=True)
    m = sym.Symbol(f"m_{name}", real=True)
    theta_t = sym.Symbol(f"theta_t_{name}", real=True)
    v_dc = sym.Symbol(f"v_dc_{name}", real=True)
    Dp_e_ref = sym.Symbol(f"Dp_e_ref_{name}", real=True)
    u_dummy = sym.Symbol(f"u_dummy_{name}", real=True)
      
    # dynamic states

    # algebraic states
    i_si = sym.Symbol(f"i_si_{name}", real=True)
    i_sr = sym.Symbol(f"i_sr_{name}", real=True)       
    p_s = sym.Symbol(f"p_s_{name}", real=True)
    q_s = sym.Symbol(f"q_s_{name}", real=True)
    p_s_ref = sym.Symbol(f"p_s_ref_{name}", real=True)
    nu_w = sym.Symbol(f"nu_w_{name}", real=True)
    
    # parameters
    S_n = sym.Symbol(f"S_n_{name}", real=True)
    F_n = sym.Symbol(f"F_n_{name}", real=True)            
    X_s = sym.Symbol(f"X_s_{name}", real=True)
    R_s = sym.Symbol(f"R_s_{name}", real=True)
    A_l = sym.Symbol(f"A_l_{name}", real=True)
    B_l = sym.Symbol(f"B_l_{name}", real=True)
    C_l = sym.Symbol(f"C_l_{name}", real=True)

    # AUXILIAR
    Omega_b = 2*np.pi*F_n
    grid.dae['params_dict'].update({f"S_n_{name}":data_dict['S_n']})

    ## PV model
    I_sc,I_mpp,V_mpp,V_oc = sym.symbols(f'I_sc_{name},I_mpp_{name},V_mpp_{name},V_oc_{name}', real=True)
    N_s,K_vt,K_it = sym.symbols(f'N_s_{name},K_vt_{name},K_it_{name}', real=True)
    K_d,R_pv_s,R_pv_sh = sym.symbols(f'K_d_{name},R_pv_s_{name},R_pv_sh_{name}', real=True)

    temp_deg,irrad,i_pv,v_pv,p_pv = sym.symbols(f'temp_deg_{name},irrad_{name},i_pv_{name},v_pv_{name},p_pv_{name}', real=True)
    
    T_stc = 25 + 273.4
    E_c = 1.6022e-19 # Elementary charge
    Boltzmann = 1.3806e-23 # Boltzmann constant

    params_dict = {str(I_sc):data_dict['I_sc'],
                   str(I_mpp):data_dict['I_mpp'],
                   str(V_mpp):data_dict['V_mpp'],
                   str(V_oc):data_dict['V_oc'],
                   str(N_s):data_dict['N_s'],
                   str(K_vt):data_dict['K_vt'],
                   str(K_it):data_dict['K_it'],
                   str(R_pv_s):data_dict['R_pv_s'],
                   str(R_pv_sh):data_dict['R_pv_sh'],
                   str(K_d):data_dict['K_d'],
                   }
    
    temp_k = temp_deg +273.4

    N_ms,N_mp = sym.symbols(f'N_ms_{name},N_mp_{name}', real=True)

    I_rrad_sts = 1000

    S_n = data_dict['S_n']
    V_dc_b = data_dict['U_n']*np.sqrt(2)
    I_dc_b = S_n/V_dc_b
    # N_ms = data_dict['N_ms'] 
    # N_mp = data_dict['N_mp'] 
    v_pv = v_dc*V_dc_b/N_ms

    grid.dae['params_dict'].update({str(N_ms):data_dict['N_ms'],str(N_mp):data_dict['N_mp'] })


    # I_sc_t = I_sc*irrad/I_rrad_sts*(1 + K_it/100*(temp_k - T_stc))
    # I_0 = (I_sc - (V_oc_t - I_sc_t*R_pv_s)/R_pv_sh)*sym.exp(-V_oc_t/(N_s*V_t))

    # I_ph = (I_0*sym.exp(V_oc_t/(N_s*V_t)) + V_oc_t/R_pv_sh)*irrad/I_rrad_sts

    # eq_i_pv = -i_pv + I_ph - I_0 * (sym.exp((v_pv + i_pv*R_pv_s)/(N_s*V_t))-1)-(v_pv+i_pv*R_pv_s)/R_pv_sh 

    V_t = K_d*Boltzmann*T_stc/E_c
    V_oc_t = V_oc * (1+K_vt/100.0*( temp_k - T_stc))

    I_sc_t = I_sc*(1 + K_it/100*(temp_k - T_stc))
    I_0 = (I_sc_t - (V_oc_t - I_sc_t*R_pv_s)/R_pv_sh)*sym.exp(-V_oc_t/(N_s*V_t))
    I_d = I_0*(sym.exp((v_pv+i_pv*R_pv_s)/(V_t*N_s))-1)
    I_ph = I_sc_t*irrad/I_rrad_sts
    
    eq_i_pv = -i_pv + I_ph - I_d - (v_pv+i_pv*R_pv_s)/R_pv_sh 

    grid.dae['h_dict'].update({f"v_pv_{name}":v_pv})
    grid.dae['h_dict'].update({f"p_pv_{name}":v_pv*i_pv})

    grid.dae['u_ini_dict'].update({f'irrad_{name}':1000,f'temp_deg_{name}':25})  # input for the initialization problem
    grid.dae['u_run_dict'].update({f'irrad_{name}':1000,f'temp_deg_{name}':25})  # input for the running problem, its value is updated


    grid.dae['g'] += [eq_i_pv]
    grid.dae['y_ini'] += [  i_pv]  
    grid.dae['y_run'] += [  i_pv]  
    
    grid.dae['params_dict'].update(params_dict)

    grid.dae['xy_0_dict'].update({str(i_pv):data_dict['I_sc']})


    ## grid side VSC control  
    i_sd_pq_ref,i_sq_pq_ref,v_td_ref,v_tq_ref = sym.symbols(f'i_sd_pq_ref_{name},i_sq_pq_ref_{name},v_td_ref_{name},v_tq_ref_{name}', real=True)
    v_dc_ref,q_s_ref = sym.symbols(f'v_dc_ref_{name},q_s_ref_{name}', real=True)
    K_pdc,K_idc = sym.symbols(f'K_pdc_{name},K_idc_{name}', real=True)
    omega_coi = sym.symbols(f'omega_coi_{name}', real=True)
    mode,i_sd_i_ref,i_sq_i_ref = sym.symbols(f'mode_{name},i_sd_i_ref_{name},i_sq_i_ref_{name}', real=True)
    p_ppc_ref,q_ppc_ref = sym.symbols(f'p_ppc_ref_{name},q_ppc_ref_{name}', real=True)

    delta = theta_s # ideal PLL
    v_sD = V_s*sin(theta_s)  # v_si   e^(-j)
    v_sQ = V_s*cos(theta_s)  # v_sr
    v_sd = v_sD * cos(delta) - v_sQ * sin(delta)   
    v_sq = v_sD * sin(delta) + v_sQ * cos(delta)
    omega_s = omega_coi
    v_tD_ref = v_td_ref * cos(delta) + v_tq_ref * sin(delta)   
    v_tQ_ref =-v_td_ref * sin(delta) + v_tq_ref * cos(delta)    
    v_ti_ref = v_tD_ref
    v_tr_ref = v_tQ_ref   
    m_ref = sym.sqrt(v_tr_ref**2 + v_ti_ref**2)/v_dc
    theta_t_ref = sym.atan2(v_ti_ref,v_tr_ref) 
    
    i_sd_ref = i_sd_i_ref + i_sd_pq_ref
    i_sq_ref = i_sq_i_ref + i_sq_pq_ref

    p_s_vdc_ref = - K_pdc*(v_dc_ref - v_dc) 
    
    eq_p_s_ref = -p_s_ref + sym.Piecewise((p_ppc_ref,p_ppc_ref<p_s_vdc_ref),(p_s_vdc_ref,True))
    eq_i_sd_pq_ref  = i_sd_pq_ref*v_sd + i_sq_pq_ref*v_sq - p_s_ref  
    eq_i_sq_pq_ref  = i_sq_pq_ref*v_sd - i_sd_pq_ref*v_sq - q_s_ref - q_ppc_ref
    eq_v_td_ref  = v_td_ref - R_s*i_sd_ref - X_s*i_sq_ref - v_sd  
    eq_v_tq_ref  = v_tq_ref - R_s*i_sq_ref + X_s*i_sd_ref - v_sq 

    grid.dae['g'] += [eq_p_s_ref,eq_i_sd_pq_ref,eq_i_sq_pq_ref,eq_v_td_ref,eq_v_tq_ref]
    grid.dae['y_ini'] += [p_s_ref, i_sd_pq_ref, i_sq_pq_ref, v_td_ref, v_tq_ref]  
    grid.dae['y_run'] += [p_s_ref, i_sd_pq_ref, i_sq_pq_ref, v_td_ref, v_tq_ref] 
    grid.dae['u_ini_dict'].update({f'v_dc_ref_{name}':1.5})
    grid.dae['u_run_dict'].update({f'v_dc_ref_{name}':1.5})
    grid.dae['u_ini_dict'].update({f'q_s_ref_{name}':0.0})
    grid.dae['u_run_dict'].update({f'q_s_ref_{name}':0.0})
    grid.dae['params_dict'].update({f'K_pdc_{name}':data_dict['K_pdc']}) 
    grid.dae['u_ini_dict'].update({f'mode_{name}':2})
    grid.dae['u_run_dict'].update({f'mode_{name}':2})
    grid.dae['u_ini_dict'].update({f'i_sd_i_ref_{name}':0})
    grid.dae['u_run_dict'].update({f'i_sd_i_ref_{name}':0})
    grid.dae['u_ini_dict'].update({f'i_sq_i_ref_{name}':0})
    grid.dae['u_run_dict'].update({f'i_sq_i_ref_{name}':0})

    grid.dae['u_ini_dict'].update({f'p_ppc_ref_{name}':2})
    grid.dae['u_run_dict'].update({f'p_ppc_ref_{name}':2})
    grid.dae['u_ini_dict'].update({f'q_ppc_ref_{name}':0})
    grid.dae['u_run_dict'].update({f'q_ppc_ref_{name}':0})

    grid.dae['xy_0_dict'].update({str(i_sr):0.1})
    grid.dae['h_dict'].update({f"i_sd_ref_{name}":i_sd_ref})
    grid.dae['h_dict'].update({f"i_sq_ref_{name}":i_sq_ref})
    grid.dae['h_dict'].update({f"p_s_vdc_ref_{name}":i_sq_ref})

    ## grid side VSC
    v_sr =  V_s*cos(theta_s)  # v_Q
    v_si =  V_s*sin(theta_s)  # v_D, e^(-j)
    i_tr = i_sr
    i_ti = i_si
    v_tr = v_tr_ref
    v_ti = v_ti_ref
    i_tdc = (i_ti*v_ti + i_tr*v_tr)/v_dc
    g_i_si = v_ti - R_s*i_si - X_s*i_sr - v_si  
    g_i_sr = v_tr - R_s*i_sr + X_s*i_si - v_sr 
    g_p_s  =  i_sr*v_sr + i_si*v_si - p_s  
    g_q_s  =  i_sr*v_si - i_si*v_sr - q_s 

    grid.dae['g'] += [g_i_si,g_i_sr,g_p_s,g_q_s]
    grid.dae['y_ini'] += [  i_si,  i_sr,  p_s,  q_s]  
    grid.dae['y_run'] += [  i_si,  i_sr,  p_s,  q_s]  
    
    grid.dae['u_ini_dict'].update({f'v_dc_ref_{name}':1.2})
    grid.dae['u_run_dict'].update({f'v_dc_ref_{name}':1.2})

    grid.dae['u_ini_dict'].update({f'{Dp_e_ref}':0.0})
    grid.dae['u_run_dict'].update({f'{Dp_e_ref}':0.0})

    grid.dae['u_ini_dict'].update({f'{u_dummy}':0.0})
    grid.dae['u_run_dict'].update({f'{u_dummy}':0.0})

    grid.dae['params_dict'].update({f"R_s_{name}":data_dict['R_s']}) 
    grid.dae['params_dict'].update({f"X_s_{name}":data_dict['X_s']}) 

    grid.dae['h_dict'].update({f"{str(p_s)}":p_s})
    grid.dae['h_dict'].update({f"{str(q_s)}":q_s}) 
    grid.dae['h_dict'].update({f"{str(i_si)}":i_si})
    grid.dae['h_dict'].update({f"{str(i_sr)}":i_sr})
    grid.dae['h_dict'].update({f"i_tdc_{name}":i_tdc})
    grid.dae['h_dict'].update({f"m_ref_{name}":m_ref})
    grid.dae['h_dict'].update({f"theta_t_ref_{name}":theta_t_ref}) 

    ## DC link   
    C_dc,i_dc = sym.symbols(f'C_dc_{name},i_dc_{name}', real=True)
    i_tdc = (i_ti*v_ti + i_tr*v_tr)/(v_dc + 1e-6)
    i_pv_pu = i_pv/I_dc_b*N_mp
    dv_dc = 0.5*(i_pv_pu - i_tdc)/(C_dc)

    grid.dae['f'] += [dv_dc]
    grid.dae['x'] += [ v_dc]
    grid.dae['params_dict'].update({f"C_dc_{name}":data_dict['C_dc']}) 
    grid.dae['xy_0_dict'].update({str(v_dc):1.2})
    grid.dae['h_dict'].update({f"i_pv_pu_{name}":i_pv_pu})
    grid.dae['h_dict'].update({f"i_pv_total_{name}":i_pv*N_mp})

    if 'A_l' in data_dict:
        grid.dae['params_dict'].update({f"A_l_{name}":data_dict['A_l']}) 
        grid.dae['params_dict'].update({f"B_l_{name}":data_dict['B_l']}) 
        grid.dae['params_dict'].update({f"C_l_{name}":data_dict['C_l']}) 
    else:
        grid.dae['params_dict'].update({f"A_l_{name}":0.005}) 
        grid.dae['params_dict'].update({f"B_l_{name}":0.005}) 
        grid.dae['params_dict'].update({f"C_l_{name}":0.005})         

    p_W   = p_s * S_n
    q_var = q_s * S_n
    return p_W,q_var


